import numpy as np
import inspect # Used for storing the input


class Element:
    def __init__(self, model, nparam, nunknowns, layers, name, label):
        self.model = model
        self.aq = None # Set in the initialization function
        self.nparam = nparam
        self.nunknowns = nunknowns
        self.layers = np.atleast_1d(layers)
        self.nlayers = len(self.layers)
        self.ncp = 0
        self.name = name
        self.label = label
        self.inhomelement = False  # elements used as part of an inhom boundary are tagged
        if self.label is not None:
            assert self.label not in list(self.model.elementdict.keys()), \
                "timml error: label " + self.label + " already exists"

    def initialize(self):
        # must be overloaded
        pass

    def potinf(self, x, y, aq=None):
        '''Returns array of size (nparam, naq)'''
        raise Exception('Must overload Element.potinf()')

    def potential(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        return np.sum(self.parameters * self.potinf(x, y, aq), 0)

    def potinflayers(self, x, y, layers, aq=None):
        '''Returns array of size (len(layers),nparam)
        only used in building equations'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        pot = self.potinf(x, y, aq)  # nparam rows, naq cols
        rv = np.sum(pot[:,np.newaxis,:] * aq.eigvec, 2).T  # Transopose as the first axes needs to be the number of layers
        return rv[layers,:]

    def potentiallayers(self, x, y, layers, aq=None):
        '''Returns array of size len(layers)
        only used in building equations'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        pot = np.sum(self.potential(x, y, aq) * aq.eigvec, 1 )
        return pot[layers]

    def disvecinf(self, x, y, aq=None):
        '''Returns array of size (2, nparam, naq)'''
        raise Exception('Must overload Element.disvecinf()')

    def disvec(self, x, y, aq=None):
        '''Returns array of size (2, nparam, naq)'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        return np.sum(self.parameters * self.disvecinf(x, y, aq), 1)

    def disvecinflayers(self, x, y, layers, aq=None):
        '''Returns two arrays of size (len(layers),nparam)
        only used in building equations'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        qxqy = self.disvecinf(x, y, aq)  # nparam rows, naq cols
        qx = np.sum(qxqy[0,:,np.newaxis,:] * aq.eigvec, 2).T  # Transpose as the first axes needs to be the number of layers
        qy = np.sum(qxqy[1,:,np.newaxis,:] * aq.eigvec, 2).T
        return np.array((qx[layers], qy[layers]))

    def disveclayers(self, x, y, layers, aq=None):
        '''Returns two arrays of size len(layers)
        only used in building equations'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        qxqy = self.disvec(x, y, aq)
        rv = np.sum(qxqy[:,np.newaxis,:] * aq.eigvec, 2)
        return rv[:,layers]

    def intpot(self, func, x1, y1, x2, y2, layers, aq=None):
        if aq is None: print('error, aquifer needs to be given')
        z1 = x1 + 1j * y1
        z2 = x2 + 1j * y2
        z = 0.5 * self.Xleg * (z2 - z1) + 0.5 * (z1 + z2)
        x = z.real
        y = z.imag
        pot = 0.0
        for i in range(self.ndeg):
            pot += self.wleg[i] * func(x=x[i], y=y[i], layers=layers, aq=aq)
        return pot

    def intflux(self, func, x1, y1, x2, y2, layers, aq=None):
        if aq is None: print('error, aquifer needs to be given')
        thetaNormOut = np.arctan2(y2 - y1, x2 - x1) - np.pi/2.0
        cosnorm = np.cos(thetaNormOut)
        sinnorm = np.sin(thetaNormOut)
        z1 = x1 + 1j * y1
        z2 = x2 + 1j * y2
        z = 0.5 * self.Xleg * (z2 - z1) + 0.5 * (z1 + z2)
        x = z.real
        y = z.imag
        qtot = 0.0
        for i in range(self.ndeg):
            qxqy = func(x=x[i], y=y[i], layers=layers, aq=aq)
            qtot += self.wleg[i] * (qxqy[0] * cosnorm + qxqy[1] * sinnorm)
        return qtot

    def headinside(self):
        print('headinside not implemented for this element')

    def setparams(self, sol):
        raise Exception('Must overload Element.setparams()')

    def storeinput(self,frame):
        self.inputargs, _, _, self.inputvalues = inspect.getargvalues(frame)

    #def stoptrace(self, xyz, layer, ltype, step, direction):
    #    return False, 0

    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer, direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        return changed, terminate, xyztnew, message

    def qztop(self, x, y):
        # given flux at top of aquifer system (as for area-sinks)
        return 0

    def plot(self, layer):
        pass
    
    def write(self):
        rv = self.name + '(' + self.model.modelname + ',\n'
        for key in self.inputargs[2:]:  # The first two are ignored
            if isinstance(self.inputvalues[key], np.ndarray):
                rv += key + ' = ' + np.array2string(self.inputvalues[key],
                                                    separator=',') + ',\n'
            elif isinstance(self.inputvalues[key], str):                
                rv += key + " = '" + self.inputvalues[key] + "',\n"
            else:
                rv += key + ' = ' + str(self.inputvalues[key]) + ',\n'
        rv += ')\n'
        return rv



import numpy as np
import inspect  # Used for storing the input
from .element import Element

__all__ = ['StripAreaSink']

class StripAreaSinkInhom(Element):
    """Create a strip area-sink in combination with an inhomogeneity.
    Created automatically using StripInhomMaq or StripInhom3D.
    Can only be created if top boundary is conf.

    Parameters
    ----------
    model : Model object
        model to which the element is added
    xleft : float
        left boundary of inhomogeneity (may not be -np.inf)
    xright : float
        right boundary of inhomogeneity (may not be np.inf)

    """
    def __init__(self, model, xleft=-1, xright=1, N=0.001, layer=0, name='StripAreaSink', label=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layer, \
                         name=name, label=label)
        self.xleft = xleft
        self.xright = xright
        self.N = N
        self.model.add_element(self)

    def __repr__(self):
        return self.name + ' between ' + str((self.xleft, self.xright))

    def initialize(self):
        self.xc = 0.5 * (self.xleft + self.xright)
        self.yc = 0
        self.L = self.xright - self.xleft
        self.aq = self.model.aq.find_aquifer_data(self.xc, self.yc)
        self.aq.add_element(self)
        self.parameters = np.array([[self.N]])
        if self.aq.ilap:
            self.lab = self.aq.lab[1:]
            self.A = -self.aq.coef[self.layers, 1:] * self.lab ** 2 / 2
            self.B = self.A * (np.exp(-self.L / self.lab) - 1)
            self.plabsq = self.aq.coef[self.layers, 1:] * self.lab ** 2
        else:
            print('StripAreaSink cannot be added to semi-confined system')

    def potinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            if (x > self.xleft) and (x < self.xright):
                rv[0, 0] = -0.5 * (x ** 2 - 2 * self.xc * x + self.xc ** 2)
                rv[0, 1:] = self.plabsq
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            if (x > self.xleft) and (x < self.xright):
                rv[0, 0, 0] = x - self.xc
        return rv

    def qztop(self, x, y):
        rv = 0.0
        if (x > self.xleft) and (x < self.xright):
            rv = -self.parameters[0, 0]  # minus cause the parameter is the infiltration rate
        return rv

    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer, direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        eps = 1e-8
        r1sq = (xyzt1[0] - self.xc) ** 2 + (xyzt1[1] - self.yc) ** 2
        r2sq = (xyzt2[0] - self.xc) ** 2 + (xyzt2[1] - self.yc) ** 2
        if (r1sq < self.Rsq and r2sq > self.Rsq ) or (r1sq > self.Rsq and r2sq < self.Rsq):
            changed = True
            x1, y1 = xyzt1[0:2]
            x2, y2 = xyzt2[0:2]
            a = (x2 - x1) ** 2 + (y2 - y1) ** 2
            b = 2 * ((x2 - x1) * (x1 - self.xc) + (y2 - y1) * (y1 - self.yc))
            c = self.xc ** 2 + self.yc ** 2 + x1 ** 2 + y1 ** 2 - 2 * (self.xc * x1 +self.yc * y1) - self.Rsq
            u1 = (-b - np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            u2 = (-b + np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            if u1 > 0:
                u = u1 * (1.0 + eps) # Go just beyond circle
            else:
                u = u2 * (1.0 + eps) # Go just beyond circle
            xn = x1 + u * (x2 - x1)
            yn = y1 + u * (y2 - y1)
            zn = xyzt1[2] + u * (xyzt2[2] - xyzt1[2])
            xyztnew = xyzt1 + u * (xyzt2 - xyzt1)
        return changed, terminate, xyztnew, message

class StripAreaSink(Element):
    def __init__(self, model, xleft=-1, xright=1, N=0.001, layer=0, name='StripAreaSink', label=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layer, \
                         name=name, label=label)
        self.xleft = xleft
        self.xright = xright
        self.N = N
        self.model.add_element(self)

    def __repr__(self):
        return self.name + ' at ' + str((self.xc, self.yc))

    def initialize(self):
        self.xc = 0.5 * (self.xleft + self.xright)
        self.yc = 0
        self.L = self.xright - self.xleft
        self.aq = self.model.aq.find_aquifer_data(self.xc, self.yc)
        self.aq.add_element(self)
        self.parameters = np.array([[self.N]])
        if self.aq.ilap:
            self.lab = self.aq.lab[1:]
            self.A = -self.aq.coef[self.layers, 1:] * self.lab ** 2 / 2
            self.B = self.A * (np.exp(-self.L / self.lab) - 1)
            self.plabsq = self.aq.coef[self.layers, 1:] * self.lab ** 2
        else:
            print('StripAreaSink cannot be added to semi-confined system')

    def potinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            if x < self.xleft:
                rv[0, 0] = -(self.xleft - self.xc) * (x - self.xc) + \
                           self.L ** 2 / 8
                rv[0, 1:] = self.B * np.exp((x - self.xleft) / self.lab)
            elif x > self.xright:
                rv[0, 0] = -(self.xright - self.xc) * (x - self.xc) + \
                           self.L ** 2 / 8
                rv[0, 1:] = self.B * np.exp(-(x - self.xright) / self.lab)
            else:
                rv[0, 0] = -0.5 * (x ** 2 - 2 * self.xc * x + self.xc ** 2)
                rv[0, 1:] = self.A * (np.exp(-(x - self.xleft) / self.lab) +
                                      np.exp((x - self.xright) / self.lab)) + \
                                      self.plabsq
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            if x < self.xleft:
                rv[0, 0, 0] = self.xleft - self.xc
                rv[0, 0, 1:] = -self.B / self.lab * np.exp((x - self.xleft) / self.lab)
            elif x > self.xright:
                rv[0, 0, 0] = self.xright - self.xc
                rv[0, 0, 1:] = self.B / self.lab * np.exp(-(x - self.xright) / self.lab)
            else:
                rv[0, 0, 0] = x - self.xc
                rv[0, 0, 1:] = self.A / self.lab * (
                                   np.exp(-(x - self.xleft) / self.lab) -
                                   np.exp((x - self.xright) / self.lab))
        return rv

    def qztop(self, x, y):
        rv = 0.0
        if (x > self.xleft) and (x < self.xright):
            rv = -self.parameters[0, 0]  # minus cause the parameter is the infiltration rate
        return rv

    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer, direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        eps = 1e-8
        r1sq = (xyzt1[0] - self.xc) ** 2 + (xyzt1[1] - self.yc) ** 2
        r2sq = (xyzt2[0] - self.xc) ** 2 + (xyzt2[1] - self.yc) ** 2
        if (r1sq < self.Rsq and r2sq > self.Rsq ) or (r1sq > self.Rsq and r2sq < self.Rsq):
            changed = True
            x1, y1 = xyzt1[0:2]
            x2, y2 = xyzt2[0:2]
            a = (x2 - x1) ** 2 + (y2 - y1) ** 2
            b = 2 * ((x2 - x1) * (x1 - self.xc) + (y2 - y1) * (y1 - self.yc))
            c = self.xc ** 2 + self.yc ** 2 + x1 ** 2 + y1 ** 2 - 2 * (self.xc * x1 +self.yc * y1) - self.Rsq
            u1 = (-b - np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            u2 = (-b + np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            if u1 > 0:
                u = u1 * (1.0 + eps) # Go just beyond circle
            else:
                u = u2 * (1.0 + eps) # Go just beyond circle
            xn = x1 + u * (x2 - x1)
            yn = y1 + u * (y2 - y1)
            zn = xyzt1[2] + u * (xyzt2[2] - xyzt1[2])
            xyztnew = xyzt1 + u * (xyzt2 - xyzt1)
        return changed, terminate, xyztnew, message



import numpy as np
import matplotlib.pyplot as plt
import inspect  # Used for storing the input
from .element import Element
from .besselaesnumba import besselaesnumba
besselaesnumba.initialize()
try:
    from .src import besselaesnew
    besselaesnew.besselaesnew.initialize()
    #print('succes on f2py')
except:
    pass
from .controlpoints import controlpoints
from .equation import DisvecEquation, LeakyWallEquation

__all__ = ['ImpLineDoublet', 'ImpLineDoubletString', 'LeakyLineDoublet', 
           'LeakyLineDoubletString']


class LineDoubletHoBase(Element):
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, delp=0.0, res=0.0,
                 layers=0, order=0, name='LineDoubletHoBase',
                 label=None, addtomodel=True, aq=None, zcinout=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers,
                         name=name, label=label)
        self.x1 = float(x1)
        self.y1 = float(y1)
        self.x2 = float(x2)
        self.y2 = float(y2)
        self.delp = np.atleast_1d(delp).astype('d')
        self.res = np.atleast_1d(res).astype('d')
        self.order = order
        self.nparam = self.nlayers * (self.order + 1)
        self.addtomodel = addtomodel
        if addtomodel: self.model.add_element(self)
        self.aq = aq
        self.zcinout = zcinout
        if self.model.f2py:
            self.bessel = besselaesnew.besselaesnew
        else:
            self.bessel = besselaesnumba
            
    def __repr__(self):
        return self.name + ' from ' + str((self.x1, self.y1)) + ' to ' + str(
            (self.x2, self.y2))

    def initialize(self):
        self.ncp = self.order + 1
        self.z1 = self.x1 + 1j * self.y1
        self.z2 = self.x2 + 1j * self.y2
        self.L = np.abs(self.z1 - self.z2)
        self.thetaNormOut = np.arctan2(self.y2 - self.y1,
                                       self.x2 - self.x1) - np.pi / 2.0
        self.cosnorm = np.cos(self.thetaNormOut) * np.ones(self.ncp)
        self.sinnorm = np.sin(self.thetaNormOut) * np.ones(self.ncp)
        #
        self.xc, self.yc = controlpoints(self.ncp, self.z1, self.z2, eps=0)
        if self.zcinout is not None:
            self.xcin, self.ycin = controlpoints(self.ncp, self.zcinout[0],
                                                 self.zcinout[1], eps=0)
            self.xcout, self.ycout = controlpoints(self.ncp, self.zcinout[2],
                                                   self.zcinout[3], eps=0)
        else:
            self.xcin, self.ycin = controlpoints(self.ncp, self.z1, self.z2,
                                                 eps=1e-6)
            self.xcout, self.ycout = controlpoints(self.ncp, self.z1, self.z2,
                                                   eps=-1e-6)
        if self.aq is None:
            self.aq = self.model.aq.find_aquifer_data(self.xc, self.yc)
        self.resfac = self.aq.T[self.layers] / self.res
        if self.addtomodel:
            self.aq.add_element(self)
        self.parameters = np.empty((self.nparam, 1))
        # Not sure if this needs to be here
        self.parameters[:, 0] = self.delp

    def potinf(self, x, y, aq=None):
        '''Can be called with only one x,y value
        Returns array(nparam, self.aq.naq) with order
        order 0, layer[0]
        order 0, layer[1]
        ...
        order 1, layer[0]
        order 1, layer[1]
        etc
        '''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            potrv = rv.reshape((self.order + 1, self.nlayers,
                                aq.naq))  # clever way of using a reshaped rv here
            pot = np.zeros((self.order + 1, aq.naq))
            pot[:, :] = self.bessel.potbesldv(float(x), float(y), self.z1, self.z2, aq.lab,
                                              self.order, aq.ilap, aq.naq)
            potrv[:] = self.aq.coef[self.layers] * pot[:, np.newaxis, :]
        return rv

    def disvecinf(self, x, y, aq=None):
        '''Can be called with only one x,y value
        Returns array(nparam, self.aq.naq) with order
        order 0, layer[0]
        order 0, layer[1]
        ...
        order 1, layer[0]
        order 1, layer[1]
        etc
        '''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            qxqyrv = rv.reshape((2, self.order + 1, self.nlayers,
                                 aq.naq))  # clever way of using a reshaped rv here
            qxqy = np.zeros((2 * (self.order + 1), aq.naq))
            qxqy[:, :] = self.bessel.disbesldv(float(x), float(y), self.z1, self.z2, aq.lab,
                                               self.order, aq.ilap, aq.naq)
            qxqyrv[0, :] = self.aq.coef[self.layers] * qxqy[:self.order + 1,
                                                       np.newaxis, :]
            qxqyrv[1, :] = self.aq.coef[self.layers] * qxqy[self.order + 1:,
                                                       np.newaxis, :]
        return rv

    def plot(self, layer=None):
        if (layer is None) or (layer in self.layers):
            plt.plot([self.x1, self.x2], [self.y1, self.y2], 'k')

class ImpLineDoublet(LineDoubletHoBase, DisvecEquation):
    """
    Create a segment of an impermeable wall, which is
    simulated with a line-doublet
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    x1 : scalar
        x-coordinate of fist point of line-doublet
    y1 : scalar
        y-coordinate of fist point of line-doublet
    x2 : scalar
        x-coordinate of second point of line-doublet
    y2 : scalar
        y-coordinate of second point of line-doublet
    order : int (default is 0)
        polynomial order of potential jump along line-doublet
        (head jump if transmissivity is equal on each side of wall)
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element
    
    See Also
    --------
    
    :class:`.ImpLineDoubletString`
    
    """
    
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, \
                 order=0, layers=0, label=None, addtomodel=True):
        self.storeinput(inspect.currentframe())
        LineDoubletHoBase.__init__(self, model, x1, y1, x2, y2, delp=0, \
                                res = np.inf, layers=layers, order=order,
                                name='ImpLineDoublet', label=label, \
                                addtomodel=addtomodel)
        self.nunknowns = self.nparam

    def initialize(self):
        LineDoubletHoBase.initialize(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol
        
class LeakyLineDoublet(LineDoubletHoBase, LeakyWallEquation):
    """
    Create a segment of a leaky wall, which is
    simulated with a line-doublet. The specific discharge through
    the wall is equal to the head difference across the wall
    divided by the resistance of the wall. 
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    x1 : scalar
        x-coordinate of fist point of line-doublet
    y1 : scalar
        y-coordinate of fist point of line-doublet
    x2 : scalar
        x-coordinate of second point of line-doublet
    y2 : scalar
        y-coordinate of second point of line-doublet
    res : scalar
        resistance of leaky wall
    order : int (default is 0)
        polynomial order of potential jump along line-doublet
        (head jump if transmissivity is equal on each side of wall)
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element
    
    See Also
    --------
    
    :class:`.LeakyLineDoubletString`
    
    """
    
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, res=0,\
                 order=0, layers=0, label=None, addtomodel=True):
        self.storeinput(inspect.currentframe())
        LineDoubletHoBase.__init__(self, model, x1, y1, x2, y2, delp=0, \
                                res=res, layers=layers, order=order,
                                name='ImpLineDoublet', label=label, \
                                addtomodel=addtomodel)
        self.nunknowns = self.nparam

    def initialize(self):
        LineDoubletHoBase.initialize(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol


class LineDoubletStringBase(Element):
    def __init__(self, model, xy, closed=False, layers=0, order=0, res=0,
                 name='LineDoubletStringBase', label=None, aq=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers, \
                         name=name, label=label)
        self.xy = np.atleast_2d(xy).astype('d')
        if closed: self.xy = np.vstack((self.xy, self.xy[0]))
        self.order = order
        self.aq = aq
        self.ldlist = []
        self.x, self.y = self.xy[:, 0], self.xy[:, 1]
        self.Nld = len(self.x) - 1
        for i in range(self.Nld):
            self.ldlist.append(
                LineDoubletHoBase(model, x1=self.x[i], y1=self.y[i], x2=self.x[i + 1],
                                  y2=self.y[i + 1], delp=0.0, res=res, layers=layers,
                                  order=order, label=label, addtomodel=False, aq=aq))

    def __repr__(self):
        return self.name + ' with nodes ' + str(self.xy)

    def initialize(self):
        for ld in self.ldlist:
            ld.initialize()
        self.ncp = self.Nld * self.ldlist[
            0].ncp  # Same order for all elements in string
        self.nparam = self.Nld * self.ldlist[0].nparam
        self.nunknowns = self.nparam
        self.xld = np.empty((self.Nld, 2))
        self.yld = np.empty((self.Nld, 2))
        for i, ld in enumerate(self.ldlist):
            self.xld[i, :] = [ld.x1, ld.x2]
            self.yld[i, :] = [ld.y1, ld.y2]
        if self.aq is None:
            self.aq = self.model.aq.find_aquifer_data(self.ldlist[0].xc,
                                                      self.ldlist[0].yc)
        self.parameters = np.zeros((self.nparam, 1))
        ## As parameters are only stored for the element not the list, we need to combine the following
        self.xc = np.array([ld.xc for ld in self.ldlist]).flatten()
        self.yc = np.array([ld.yc for ld in self.ldlist]).flatten()
        self.xcin = np.array([ld.xcin for ld in self.ldlist]).flatten()
        self.ycin = np.array([ld.ycin for ld in self.ldlist]).flatten()
        self.xcout = np.array([ld.xcout for ld in self.ldlist]).flatten()
        self.ycout = np.array([ld.ycout for ld in self.ldlist]).flatten()
        self.cosnorm = np.array([ld.cosnorm for ld in self.ldlist]).flatten()
        self.sinnorm = np.array([ld.sinnorm for ld in self.ldlist]).flatten()
        self.aqin = self.model.aq.find_aquifer_data(self.xcin[0], self.ycin[0])
        self.aqout = self.model.aq.find_aquifer_data(self.xcout[0],
                                                     self.ycout[0])
        self.resfac = self.ldlist[0].resfac

    def potinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.Nld, self.ldlist[0].nparam, aq.naq))
        for i in range(self.Nld):
            rv[i] = self.ldlist[i].potinf(x, y, aq)
        rv.shape = (self.nparam, aq.naq)
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.Nld, self.ldlist[0].nparam, aq.naq))
        for i in range(self.Nld):
            rv[:, i] = self.ldlist[i].disvecinf(x, y, aq)
        rv.shape = (2, self.nparam, aq.naq)
        return rv
    
    def plot(self, layer=None):
        if (layer is None) or (layer in self.layers):
            plt.plot(self.x, self.y, 'k')


class ImpLineDoubletString(LineDoubletStringBase, DisvecEquation):
    """
    Create a string of impermeable wall segements consisting
    of line-doublets
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of end-points of
        the segements in the string
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers
    order : int (default is 0)
        polynomial order of potential jump along line-doublet
        (head jump if transmissivity is equal on each side of wall)
    label: str or None
        label of element
    
    See Also
    --------
    
    :class:`.ImpLineDoublet`
    
    """
    
    
    def __init__(self, model, xy=[(-1, 0), (1, 0)], \
                 layers=0, order=0, label=None):
        self.storeinput(inspect.currentframe())
        LineDoubletStringBase.__init__(self, model, xy, closed=False,
                                    res=np.inf, layers=layers, order=order, \
                                    name='ImpLineDoubletString', label=label,
                                    aq=None)
        self.model.add_element(self)

    def initialize(self):
        LineDoubletStringBase.initialize(self)
        self.aq.add_element(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol
        
class LeakyLineDoubletString(LineDoubletStringBase, LeakyWallEquation):
    """
    Create a string of leaky wall segements consisting
    of line-doublets
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of end-points of
        the segements in the string
    res : scalar
        resistance of leaky wall
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers
    order : int (default is 0)
        polynomial order of potential jump along line-doublet
        (head jump if transmissivity is equal on each side of wall)
    label: str or None
        label of element
    
    See Also
    --------
    
    :class:`.ImpLineDoublet`
    
    """
    
    def __init__(self, model, xy=[(-1, 0), (1, 0)], res=np.inf,\
                 layers=0, order=0, label=None):
        self.storeinput(inspect.currentframe())
        LineDoubletStringBase.__init__(self, model, xy, closed=False,
                                    layers=layers, order=order, res=res,\
                                    name='ImpLineDoubletString', label=label,
                                    aq=None)
        self.model.add_element(self)

    def initialize(self):
        LineDoubletStringBase.initialize(self)
        self.aq.add_element(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol



import numpy as np
import inspect  # Used for storing the input
from .element import Element
from .equation import PotentialEquation

__all__ = ['Constant', 'ConstantStar']

class ConstantBase(Element, PotentialEquation):
    def __init__(self, model, xr=0, yr=0, hr=0.0, layer=0, \
                 name='ConstantBase', label=None, aq=None):
        Element.__init__(self, model, nparam=1, nunknowns=1, layers=layer, \
                         name=name, label=label)
        self.nparam = 1  # Defined here and not in Element as other elements can have multiple parameters per layers
        self.nunknowns = 0
        self.xr = xr
        self.yr = yr
        self.hr = hr
        self.aq = aq
        self.model.add_element(self)

    def __repr__(self):
        return self.name + ' at ' + str(
            (self.xr, self.yr)) + ' with head  ' + str(self.hr)

    def initialize(self):
        if self.aq is None:
            self.aq = self.model.aq.find_aquifer_data(self.xr, self.yr)
        self.aq.add_element(self)
        self.ncp = 1
        self.xc = np.array([self.xr])
        self.yc = np.array([self.yr])
        self.pc = self.hr * self.aq.T[self.layers]
        self.parameters = np.atleast_2d(self.pc)

    def potinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((1, aq.naq))
        if aq == self.aq:
            rv[0, 0] = 1
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, 1, aq.naq))
        return rv


class Constant(ConstantBase, PotentialEquation):
    """
    Specify the head at one point in the model in one layer.
    The head may only be specified in an area of the model where
    the aquifer system is confined.
    
    Parameters
    ----------
    model : Model object
        model to which the element is added
    xr : float
        x-coordinate of the point where the head is specified
    yr : float
        y-coordinate of the point where the head is specified
    hr : float
        specified head
    rw : float
        radius of the well
    layer : int
        layer where the head is specified
    label : string or None (default: None)
        label of the element
    
    """
    
    def __init__(self, model, xr=0, yr=0, hr=0.0, layer=0, label=None):
        self.storeinput(inspect.currentframe())
        ConstantBase.__init__(self, model, xr=xr, yr=yr, hr=hr, layer=layer, \
                              name='Constant', label=label)
        self.nunknowns = 1

    def initialize(self):
        ConstantBase.initialize(self)
        assert self.aq.ilap, 'Constant element added to area that is ' \
                             'semi-confined'
        self.resfac = np.zeros(1)  # required for HeadEquation
        self.strengthinf = np.zeros(1)  # required for HeadEquation

    def setparams(self, sol):
        self.parameters[:, 0] = sol


class ConstantInside(Element):
    # Sets constant at points xc, yc equal to the average of the potential of all elements at points xc, yc
    # Used for the inside of an inhomogeneity
    def __init__(self, model, xc=0, yc=0, label=None):
        Element.__init__(self, model, nparam=1, nunknowns=1,
                         layers=list(range(model.aq.naq)), \
                         name='ConstantInside', label=label)
        self.xc = np.atleast_1d(xc)
        self.yc = np.atleast_1d(yc)
        self.parameters = np.zeros((1, 1))
        self.model.add_element(self)

    def __repr__(self):
        return self.name

    def initialize(self):
        self.aq = self.model.aq.find_aquifer_data(self.xc[0], self.yc[0])
        self.aq.add_element(self)
        self.ncp = len(self.xc)

    def potinf(self, x, y, aq=None):
        '''Can be called with only one x,y value'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((1, aq.naq))
        if aq == self.aq:
            rv[0, 0] = 1
        return rv

    def disvecinf(self, x, y, aq=None):
        '''Can be called with only one x,y value'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, 1, aq.naq))
        return rv

    def equation(self):
        mat = np.zeros((1, self.model.neq))
        rhs = np.zeros(1)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            ieq = 0
            for e in self.model.elementlist:
                if e. nunknowns > 0:
                    if e != self:
                        mat[0:, ieq:ieq + e. nunknowns] += \
                            e.potinflayers(self.xc[icp], self.yc[icp],
                                           self.layers).sum(0)
                        ieq += e. nunknowns
                        # else:
                        #    mat[0, ieq:ieq+e. nunknowns] += -1
                else:
                    rhs[0] -= \
                        e.potentiallayers(self.xc[icp], self.yc[icp],
                                          self.layers).sum(0)
        return mat, rhs

    def setparams(self, sol):
        self.parameters[:, 0] = sol


#class ConstantStar(Element, PotentialEquation):
# I don't think we need the equation
class ConstantStar(Element):
    def __init__(self, model, hstar=0.0, label=None, aq=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=0, \
                         name='ConstantStar', label=label)
        assert hstar is not None, 'a value for hstar needs to be specified'
        self.hstar = hstar
        self.aq = aq
        self.model.add_element(self)

    def __repr__(self):
        return self.name + ' with head  ' + str(self.hstar)

    def initialize(self):
        self.aq.add_element(self)
        self.aq.constantstar = self
        self.parameters = np.zeros((1, 1))
        self.potstar = self.hstar * self.aq.T

    def potinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((1, aq.naq))
        return rv

    def potentiallayers(self, x, y, layers, aq=None):
        '''Returns array of size len(layers) only used in building equations
        Defined here as it is the particular solution inside a semi-confined aquifer
        and cannot be added by using eigen vectors'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        pot = np.zeros(len(layers))
        if aq == self.aq:
            pot[:] = self.potstar[layers]
        return pot

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, 1, aq.naq))
        return rv



import numpy as np
import inspect  # Used for storing the input
from .element import Element
from .equation import HeadEquation, MscreenWellEquation, HeadDiffEquation, DisvecDiffEquation

__all__ = ['LineSink1D', 'HeadLineSink1D']

class LineSink1DBase(Element):

    def __init__(self, model, xls, sigls=1, layers=0, name="LineSink1DBase", label=None, 
                 addtomodel=True, res=0, wh=1, aq=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers,
                         name=name, label=label)
        self.xls = float(xls)
        self.sigls = np.atleast_1d(sigls)
        self.res = float(res)
        self.wh = wh
        self.aq = aq
        self.addtomodel = addtomodel
        if self.addtomodel: 
            self.model.add_element(self)
        self.nparam = self.nlayers
        self.tiny = 1e-12
        
    def __repr__(self):
        return self.name + " at " + str(self.xls) + " in layers: " + str(self.layers)
    
    def initialize(self):
        self.xc = np.array([self.xls])
        self.yc = np.zeros(1)
        self.ncp = 1
        if self.aq is None:
            self.aq = self.model.aq.find_aquifer_data(self.xc[0], self.yc[0])

        if self.addtomodel:
            self.aq.add_element(self)
        self.parameters = np.empty((self.nparam, 1))
        self.parameters[:, 0] = self.sigls
        self.theta_norm_out = np.zeros(1)
        self.cosnorm = np.cos(self.theta_norm_out) * np.ones(self.ncp)
        self.sinnorm = np.sin(self.theta_norm_out) * np.ones(self.ncp)
        if self.wh == 'H':
            self.wh = self.aq.Haq[self.layers]
        elif self.wh == '2H':
            self.wh = 2.0 * self.aq.Haq[self.layers]
        elif np.isscalar(self.wh):
            self.wh = self.wh * np.ones(self.nlayers)
        self.resfac = self.aq.T[self.layers] * self.res / self.wh

    def potinf(self, x, y, aq=None):
        if aq is None: 
            aq = self.model.aq.find_aquifer_data(x, 0)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            pot = np.zeros(aq.naq)
            if aq.ilap:
                if x - self.xls < 0.:
                    pot[0] = -0.5 * (x - self.xls - 1)  # so that pot = 0.5 at x=xls
                    pot[1:] = -0.5 * aq.lab[1:] * np.exp((x - self.xls) / aq.lab[1:])
                elif x - self.xls >= 0.:
                    pot[0] = 0.5 * (x - self.xls + 1)
                    pot[1:] = -0.5 * aq.lab[1:] * np.exp(-(x - self.xls) / aq.lab[1:])
            else:
                if x - self.xls < 0.:
                    pot[:] = -0.5 * aq.lab * np.exp((x - self.xls) / aq.lab)
                elif x - self.xls >= 0.:
                    pot[:] = -0.5 * aq.lab * np.exp(-(x - self.xls) / aq.lab)
            rv[:] = self.aq.coef[self.layers] * pot
        return rv
    
    def disvecinf(self, x, y, aq=None):
        if aq is None: 
            aq = self.model.aq.find_aquifer_data(x, 0)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            qx = np.zeros(aq.naq)
            if aq.ilap:
                if x - self.xls < 0.:
                    qx[0] = 0.5
                    qx[1:] = 0.5 * np.exp((x - self.xls) / aq.lab[1:])
                elif x - self.xls >= 0.:
                    qx[0] = -0.5
                    qx[1:] = -0.5 * np.exp(-(x - self.xls) / aq.lab[1:])
            else:
                if x - self.xls < 0.:
                    qx[:] = 0.5 * np.exp((x - self.xls) / aq.lab)
                elif x - self.xls >= 0.:
                    qx[:] = -0.5 * np.exp(-(x - self.xls) / aq.lab)
            rv[0] = self.aq.coef[self.layers] * qx
        return rv
    
    def discharge(self):
        """Discharge per unit length"""
        Q = np.zeros(self.aq.naq)
        Q[self.layers] = self.parameters[:, 0]
        return Q
    
class LineSink1D(LineSink1DBase, MscreenWellEquation):
    """
    Create an infinitely long line-sink with a given discharge
    per unit length. In case the line-sink is screened in multiple
    layers, the discharge is distributed over the layers such that
    the head inside the line-sink is constant in all layers.
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xls : scalar
        x-location of line-sink
    sigls : scalar
        discharge per unit length
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the aquifer layer
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element

    """
    
    def __init__(self, model, xls=0, sigls=1, \
                 layers=0, label=None):
        self.storeinput(inspect.currentframe())
        LineSink1DBase.__init__(self, model, xls, sigls=sigls, layers=layers, \
                               name="Linesink1D", label=label, \
                               addtomodel=True, res=0, wh=1, aq=None)
        self.Qc = float(sigls)
        if self.nlayers == 1:
            self.nunknowns = 0
        else:
            self.nunknowns = self.nparam

    def initialize(self):
        LineSink1DBase.initialize(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol
    
class HeadLineSink1D(LineSink1DBase, HeadEquation):
    """
    Create an infinitely long line-sink with a given head.
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xls : scalar
        x-location of line-sink
    hls : scalar
        head in line-sink
    res : scalar (default is 0)
        resistance of line-sink
    wh : scalar or str
        distance over which water enters line-sink
        if 'H': (default) distance is equal to the thickness of the aquifer layer (when flow comes mainly from one side)
        if '2H': distance is twice the thickness of the aquifer layer (when flow comes from both sides)
        if scalar: the width of the stream that partially penetrates the aquifer layer
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element

    """
 
    def __init__(self, model, xls=0, hls=1, \
                 res=0, wh=1, layers=0, label=None):
        self.storeinput(inspect.currentframe())
        LineSink1DBase.__init__(self, model, xls, sigls=0, layers=layers, \
                               name="HeadLinesink1D", label=label, \
                               addtomodel=True, res=res, wh=wh, aq=None)
        self.hc = np.atleast_1d(hls) * np.ones(self.nlayers)
        self.nunknowns = self.nparam

    def initialize(self):
        LineSink1DBase.initialize(self)
        self.pc = self.hc * self.aq.T[self.layers]  # Needed in solving

    def setparams(self, sol):
        self.parameters[:, 0] = sol
    
class HeadDiffLineSink1D(LineSink1DBase, HeadDiffEquation):
    """HeadDiffLineSink1D for left side (xcout)
    """
    
    def __init__(self, model, xls, label=None, 
                 aq=None, aqin=None, aqout=None):
        LineSink1DBase.__init__(self, model, xls, sigls=0,
                                layers=np.arange(model.aq.naq), label=label, 
                                name='HeadDiffLineSink1D', addtomodel=True, aq=aq)
        self.inhomelement = True
        self.nunknowns = self.nparam
        self.aqin = aqin
        self.aqout = aqout
    
    def initialize(self):
        LineSink1DBase.initialize(self)
        self.xcout = self.xc + self.tiny * abs(self.xc) + self.tiny
        self.xcin = self.xc - self.tiny * abs(self.xc) - self.tiny
        self.ycout = np.zeros(1)
        self.ycin = np.zeros(1)
        if self.aqout is None:
            self.aqout = self.model.aq.find_aquifer_data(self.xcout[0], 0)
        if self.aqin is None:
            self.aqin = self.model.aq.find_aquifer_data(self.xcin[0], 0)
            
    def setparams(self, sol):
        self.parameters[:, 0] = sol
        
class FluxDiffLineSink1D(LineSink1DBase, DisvecDiffEquation):
    """HeadDiffLineSink1D for left side (xcout)
    """
    
    def __init__(self, model, xls, label=None, 
                 aq=None, aqin=None, aqout=None):
        LineSink1DBase.__init__(self, model, xls, sigls=0,
                                layers=np.arange(model.aq.naq), label=label, 
                                name='FluxDiffLineSink1D', addtomodel=True, aq=aq)
        self.inhomelement = True
        self.nunknowns = self.nparam
        self.aqin = aqin
        self.aqout = aqout
    
    def initialize(self):
        LineSink1DBase.initialize(self)
        self.xcout = self.xc - self.tiny * abs(self.xc) - self.tiny
        self.xcin = self.xc + self.tiny * abs(self.xc) + self.tiny
        self.ycout = np.zeros(1)
        self.ycin = np.zeros(1)
        if self.aqout is None:
            self.aqout = self.model.aq.find_aquifer_data(self.xcout[0], 0)
        if self.aqin is None:
            self.aqin = self.model.aq.find_aquifer_data(self.xcin[0], 0)
            
    def setparams(self, sol):
        self.parameters[:, 0] = sol



import numpy as np
import inspect  # Used for storing the input
from .aquifer_parameters import param_maq
from .constant import ConstantStar

class AquiferData:
    def __init__(self, model, kaq, c, z, npor, ltype):
        # All input variables except model should be numpy arrays
        # That should be checked outside this function
        self.model = model
        # Needed for heads
        self.kaq = np.atleast_1d(kaq)
        self.naq = len(kaq)
        self.c = np.atleast_1d(c)
        self.hstar = None
        # Needed for tracing
        self.z = np.atleast_1d(z)
        self.Hlayer = self.z[:-1] - self.z[1:]  # thickness of all layers
        self.nlayers = len(self.z) - 1
        self.npor = np.atleast_1d(npor)
        self.ltype = np.atleast_1d(ltype)
        # tag indicating whether an aquifer is Laplace (confined on top)
        if self.ltype[0] == 'a':
            self.ilap = 1
        else:
            self.ilap = 0
        #
        self.area = 1e200  # Smaller than default of ml.aq so that inhom is found
        self.layernumber = np.zeros(self.nlayers, dtype='int')
        self.layernumber[self.ltype == 'a'] = np.arange(self.naq)
        self.layernumber[self.ltype == 'l'] = np.arange(self.nlayers - self.naq)
        if self.ltype[0] == 'a':
            self.layernumber[self.ltype == 'l'] += 1  # first leaky layer below first aquifer layer
        self.zaqtop = self.z[:-1][self.ltype == 'a']
        self.zaqbot = self.z[1:][self.ltype == 'a']
        self.Haq = self.zaqtop - self.zaqbot
        self.T = self.kaq * self.Haq
        self.Tcol = self.T[:, np.newaxis]
        self.zlltop = self.z[:-1][self.ltype == 'l']
        self.zllbot = self.z[1:][self.ltype == 'l']
        if self.ltype[0] == 'a':
            self.zlltop = np.hstack((self.z[0], self.zlltop))
            self.zllbot = np.hstack((self.z[0], self.zllbot))
        self.Hll = self.zlltop - self.zllbot
        self.nporaq = self.npor[self.ltype == 'a']
        if self.ltype[0] == 'a':
            self.nporll =  np.ones(len(self.npor[self.ltype == 'l']) + 1)
            self.nporll[1:] = self.npor[self.ltype == 'l']
        else:
            self.nporll = self.npor[self.ltype == 'l']

    def initialize(self):
        self.elementlist = []  # Elementlist of aquifer
        d0 = 1.0 / (self.c * self.T)
        d0[:-1] += 1.0 / (self.c[1:] * self.T[:-1])
        dp1 = -1.0 / (self.c[1:] * self.T[1:])
        dm1 = -1.0 / (self.c[1:] * self.T[:-1])
        A = np.diag(dm1, -1) + np.diag(d0, 0) + np.diag(dp1, 1)
        w, v = np.linalg.eig(A)
        # sort lab in decending order, hence w in ascending order
        index = np.argsort(abs(w))
        w = w[index]
        v = v[:, index]
        if self.ilap:
            self.lab = np.zeros(self.naq)
            self.lab[1:] = 1.0 / np.sqrt(w[1:])
            self.zeropluslab = self.lab  # to be deprecated when new lambda is fully implemented
            v[:, 0] = self.T / np.sum(self.T)  # first column is normalized T
        else:
            self.lab = 1.0 / np.sqrt(w)
        self.eigvec = v
        self.coef = np.linalg.solve(v, np.diag(np.ones(self.naq))).T

    def add_element(self, e):
        self.elementlist.append(e)
        if isinstance(e, ConstantStar):
            self.hstar = e.hstar

    def isinside(self, x, y):
        raise Exception('Must overload AquiferData.isinside()')

    def storeinput(self, frame):
        self.inputargs, _, _, self.inputvalues = inspect.getargvalues(frame)
        
    def findlayer(self, z):
        '''
        Returns layer-number, layer-type and model-layer-number'''
        if z > self.z[0]:
            modellayer, ltype = -1, 'above'
            layernumber = None
        elif z < self.z[-1]:
            modellayer, ltype = len(self.layernumber), 'below'
            layernumber = None
        else:
            modellayer = np.argwhere((z <= self.z[:-1]) & (z >= self.z[1:]))[0, 0]
            layernumber = self.layernumber[modellayer]
            ltype = self.ltype[modellayer] 
        return layernumber, ltype, modellayer


class Aquifer(AquiferData):
    def __init__(self, model, kaq, c, z, npor, ltype):
        AquiferData.__init__(self, model, kaq, c, z, npor, ltype)
        self.inhomlist = []
        self.area = 1e300  # Needed to find smallest inhom

    def initialize(self):
        # cause we are going to call initialize for inhoms
        AquiferData.initialize(self)
        for inhom in self.inhomlist:
            inhom.initialize()
        for inhom in self.inhomlist:
            inhom.create_elements()

    def add_inhom(self, inhom):
        self.inhomlist.append(inhom)
        return len(self.inhomlist) - 1  # returns number in the list

    def find_aquifer_data(self, x, y):
        rv = self
        for inhom in self.inhomlist:
            if inhom.isinside(x, y):
                if inhom.area < rv.area:
                    rv = inhom
        return rv
        # Not used anymore I think 5 Nov 2015
        # def find_aquifer_number(self, x, y):
        #    rv = -1
        #    for i,inhom in enumerate(self.inhomlist):
        #        if inhom.isinside(x, y):
        #            if inhom.area < rv.area:
        #                rv = i
        #    return rv


        



import numpy as np
import inspect # Used for storing the input
from .element import Element

__all__ = ['Uflow']

class Uflow(Element):
    """
    Add uniform flow to the model.
    Uniform flow may only be added to a model of which the background
    aquifer system is confined.
    
    Parameters
    ----------
    model : Model object
        model to which the uniform flow is added
    slope : float
        slope of the head (head drop divided by the distance) in the
        direction of flow
    angle : float
        direction of flow in degerees (0 degrees is straight East,
        counter clock-wise is positive)
    label : string or None (default: None)
        label of the element
    
    """
    
    def __init__(self, model, slope, angle, label=None):
        assert model.aq.ilap, 'TimML Error: Uflow can only be added to model with background confined aquifer'
        self.storeinput(inspect.currentframe())
        Element.__init__(self, model, nparam=2, nunknowns=0, layers=0, \
                         name='Uflow', label=label)
        self.slope = slope
        self.angle = angle
        self.model.add_element(self)
    def __repr__(self):
        return self.name + ' with slope and angle: ' + str((self.slope, self.angle))
    def initialize(self):
        self.aq = self.model.aq
        self.aq.add_element(self)
        self.Qx = self.slope * np.cos(self.angle * np.pi / 180) * np.sum(self.aq.T)
        self.Qy = self.slope * np.sin(self.angle * np.pi / 180) * np.sum(self.aq.T)
        self.parameters = np.array([[self.Qx], [self.Qy]])
    def potinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, aq.naq))
        if aq == self.aq:
            rv[0, 0] = -x
            rv[1, 0] = -y
        return rv
    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, 2, aq.naq))
        if aq == self.aq:
            rv[0, 0, 0] = 1.0
            rv[1, 1, 0] = 1.0
        return rv



import numpy as np


class PotentialEquation:
    def equation(self):
        """Mix-in class that returns matrix rows for potential-specified conditions.
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        # rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        rhs = self.pc.copy()
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            # rhs[istart:istart+self.nlayers] = self.pc[]
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = \
                        e.potinflayers(self.xc[icp], self.yc[icp], self.layers)
                    if e == self:
                        mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] -= self.resfac[icp]
                    ieq += e.nunknowns
                else:
                    rhs[istart:istart + self.nlayers] -= \
                        e.potentiallayers(self.xc[icp], self.yc[icp],
                                          self.layers)  # Pretty cool that this works, really
        return mat, rhs


class HeadEquation:
    def equation(self):
        """Mix-in class that returns matrix rows for head-specified conditions. (now written as heads)
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        # rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        rhs = self.hc.copy()
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            # rhs[istart:istart+self.nlayers] = self.pc[]
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = \
                        e.potinflayers(self.xc[icp], self.yc[icp], self.layers) / self.aq.Tcol[self.layers]
                    if e == self:
                        mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] -= self.resfac[icp]
                    ieq += e.nunknowns
                else:
                    rhs[istart:istart + self.nlayers] -= \
                        e.potentiallayers(self.xc[icp], self.yc[icp], self.layers) / self.aq.T[
                            self.layers]  # Pretty cool that this works, really
        return mat, rhs


class HeadEquationNoRes:  # This class can be deleted when HeadEquation works with zero resistance
    def equation(self):
        """Mix-in class that returns matrix rows for head-specified conditions. (really written as constant potential element)
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            rhs[istart:istart + self.nlayers] = self.pc
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = \
                        e.potinflayers(self.xc[icp], self.yc[icp], self.layers)
                    ieq += e.nunknowns
                else:
                    rhs[istart:istart + self.nlayers] -= \
                        e.potentiallayers(self.xc[icp], self.yc[icp],
                                          self.layers)  # Pretty cool that this works, really
        return mat, rhs


class MscreenWellEquation:
    def equation(self):
        """Mix-in class that returns matrix rows for mscreen condition.
        Mscreen condition applied at each control point separately (so not like in ditch).
        Returns matrix part (nunknowns,neq)
        Returns rhs part (nunknowns)
        """
        mat = np.zeros((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        rhs[0:self.nlayers - 1] = 0.0
        rhs[self.nlayers - 1] = self.Qc
        ieq = 0
        for e in self.model.elementlist:
            if e.nunknowns > 0:
                head = e.potinflayers(self.xc, self.yc, self.layers) / self.aq.Tcol[self.layers, :]
                mat[0:self.nlayers - 1, ieq:ieq + e.nunknowns] = head[:-1] - head[1:]
                if e == self:
                    for i in range(self.nlayers - 1):
                        mat[i, ieq + i] -= self.resfac[i]
                        mat[i, ieq + i + 1] += self.resfac[i + 1]
                    mat[self.nlayers - 1, ieq:ieq + self.nlayers] = 1.0
                ieq += e.nunknowns
            else:
                head = e.potentiallayers(self.xc, self.yc, self.layers) / self.aq.T[self.layers]
                rhs[0:self.nlayers - 1] -= head[:-1] - head[1:]
        return mat, rhs


class DisvecEquation:
    def equation(self):
        """Mix-in class that returns matrix rows for zero normal flux conditions.
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    qx, qy = e.disvecinflayers(self.xc[icp], self.yc[icp], self.layers)
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = \
                        qx * self.cosnorm[icp] + qy * self.sinnorm[icp]
                    ieq += e.nunknowns
                else:
                    qx, qy = e.disveclayers(self.xc[icp], self.yc[icp], self.layers)
                    rhs[istart:istart + self.nlayers] -= qx * self.cosnorm[icp] + qy * self.sinnorm[icp]
        return mat, rhs


class DisvecEquationOut:
    def equation(self):
        """Mix-in class that returns matrix rows for zero normal flux condition.
        Using the control point on the outside
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    qx, qy = e.disvecinflayers(self.xcout[icp], self.ycout[icp], self.layers)
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = \
                        qx * self.cosnorm[icp] + qy * self.sinnorm[icp]
                    ieq += e.nunknowns
                else:
                    qx, qy = e.disveclayers(self.xcout[icp], self.ycout[icp], self.layers)
                    rhs[istart:istart + self.nlayers] -= qx * self.cosnorm[icp] + qy * self.sinnorm[icp]
        return mat, rhs


class LeakyWallEquation:
    def equation(self):
        """Mix-in class that returns matrix rows for leaky wall condition.
        Qnormal = resfac * (headin - headout)
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    qx, qy = e.disvecinflayers(self.xc[icp], self.yc[icp], self.layers)
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = \
                        qx * self.cosnorm[icp] + qy * self.sinnorm[icp] - self.resfac[:, np.newaxis] * \
                        (e.potinflayers(self.xcin[icp], self.ycin[icp], self.layers, aq=self.aq) / self.aq.Tcol[
                            self.layers] - \
                         e.potinflayers(self.xcout[icp], self.ycout[icp], self.layers, aq=self.aq) / self.aq.Tcol[
                             self.layers])
                    ieq += e.nunknowns
                else:
                    qx, qy = e.disveclayers(self.xc[icp], self.yc[icp], self.layers)
                    rhs[istart:istart + self.nlayers] -= qx * self.cosnorm[icp] + qy * self.sinnorm[icp] + self.resfac * \
                                                          (e.potentiallayers(self.xcin[icp], self.ycin[icp], self.layers,
                                                                             aq=self.aq) / self.aq.T[self.layers] -
                                                           e.potentiallayers(self.xcout[icp], self.ycout[icp],
                                                                             self.layers, aq=self.aq) / self.aq.T[
                                                              self.layers])
        return mat, rhs


class HeadDiffEquation:
    def equation(self):
        """Mix-in class that returns matrix rows for difference in head between inside and
        outside equals zeros
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = \
                        e.potinflayers(self.xcin[icp], self.ycin[icp], self.layers, aq=self.aqin) / self.aqin.Tcol - \
                        e.potinflayers(self.xcout[icp], self.ycout[icp], self.layers, aq=self.aqout) / self.aqout.Tcol
                    ieq += e.nunknowns
                else:
                    rhs[istart:istart + self.nlayers] -= \
                        e.potentiallayers(self.xcin[icp], self.ycin[icp], self.layers, aq=self.aqin) / self.aqin.T - \
                        e.potentiallayers(self.xcout[icp], self.ycout[icp], self.layers, aq=self.aqout) / self.aqout.T
        return mat, rhs


class HeadDiffEquation2:
    # Integrated head inside and outside are equal
    def equation(self):
        """Mix-in class that returns matrix rows for difference in head between inside and
        outside equals zeros
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    headin = self.intpot(e.potinflayers, self.xcin[icp], self.ycin[icp],
                                         self.xcin[icp + 1], self.ycin[icp + 1], self.layers,
                                         aq=self.aqin) / self.aqin.Tcol[self.layers]
                    headout = self.intpot(e.potinflayers, self.xcout[icp], self.ycout[icp],
                                          self.xcout[icp + 1], self.ycout[icp + 1], self.layers,
                                          aq=self.aqout) / self.aqout.Tcol[self.layers]
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = headin - headout
                    ieq += e.nunknowns
                else:
                    headin = self.intpot(e.potentiallayers, self.xcin[icp], self.ycin[icp],
                                         self.xcin[icp + 1], self.ycin[icp + 1], self.layers,
                                         aq=self.aqin) / self.aqin.T[self.layers]
                    headout = self.intpot(e.potentiallayers, self.xcout[icp], self.ycout[icp],
                                          self.xcout[icp + 1], self.ycout[icp + 1], self.layers,
                                          aq=self.aqout) / self.aqout.T[self.layers]
                    rhs[istart:istart + self.nlayers] -= headin - headout
        return mat, rhs


class DisvecDiffEquation:
    def equation(self):
        """Mix-in class that returns matrix rows for difference in head between inside and
        outside equals zeros
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    qxin, qyin = e.disvecinflayers(self.xcin[icp], self.ycin[icp], self.layers, aq=self.aqin)
                    qxout, qyout = e.disvecinflayers(self.xcout[icp], self.ycout[icp], self.layers, aq=self.aqout)
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = \
                        (qxin - qxout) * self.cosnorm[icp] + (qyin - qyout) * self.sinnorm[icp]
                    ieq += e.nunknowns
                else:
                    qxin, qyin = e.disveclayers(self.xcin[icp], self.ycin[icp], self.layers, aq=self.aqin)
                    qxout, qyout = e.disveclayers(self.xcout[icp], self.ycout[icp], self.layers, aq=self.aqout)
                    rhs[istart:istart + self.nlayers] -= (qxin - qxout) * self.cosnorm[icp] + (qyin - qyout) * \
                                                         self.sinnorm[icp]
        return mat, rhs


class DisvecDiffEquation2:
    def equation(self):
        """Mix-in class that returns matrix rows for difference in head between inside and
        outside equals zeros
        Returns matrix part (nunknowns,neq)
        Returns rhs part nunknowns
        """
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    fluxin = self.intflux(e.disvecinflayers, self.xcin[icp], self.ycin[icp],
                                          self.xcin[icp + 1], self.ycin[icp + 1], self.layers, aq=self.aqin)
                    fluxout = self.intflux(e.disvecinflayers, self.xcout[icp], self.ycout[icp],
                                           self.xcout[icp + 1], self.ycout[icp + 1], self.layers, aq=self.aqout)
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = fluxin - fluxout
                    ieq += e.nunknowns
                else:
                    fluxin = self.intflux(e.disveclayers, self.xcin[icp], self.ycin[icp],
                                          self.xcin[icp + 1], self.ycin[icp + 1], self.layers, aq=self.aqin)
                    fluxout = self.intflux(e.disveclayers, self.xcout[icp], self.ycout[icp],
                                           self.xcout[icp + 1], self.ycout[icp + 1], self.layers, aq=self.aqout)
                    rhs[istart:istart + self.nlayers] -= fluxin - fluxout
        return mat, rhs


class IntDisVecEquation:
    def equation(self):
        mat = np.empty((self.nunknowns, self.model.neq))
        rhs = np.zeros(self.nunknowns)  # Needs to be initialized to zero
        for icp in range(self.ncp):
            istart = icp * self.nlayers
            ieq = 0
            for e in self.model.elementlist:
                if e.nunknowns > 0:
                    flux = self.intflux(e.disvecinflayers, self.xc[icp], self.yc[icp],
                                        self.xc[icp + 1], self.yc[icp + 1], self.layers, aq=self.aq)
                    mat[istart:istart + self.nlayers, ieq:ieq + e.nunknowns] = flux
                    ieq += e.nunknowns
                else:
                    flux = self.intflux(e.disveclayers, self.xc[icp], self.yc[icp],
                                        self.xc[icp + 1], self.yc[icp + 1], self.layers, aq=self.aq)
                    rhs[istart:istart + self.nlayers] -= flux

        return mat, rhs



import numpy as np
import inspect  # Used for storing the input
from .element import Element
from scipy.special import k0, k1, i0, i1

__all__ = ['CircAreaSink']

class CircAreaSink(Element):
    """
    Class to create a circular area-sink.

    Parameters
    ----------
    model : Model object
        model to which the element is added
    xc : float
        x-coordinate of center of area-sink
    yc : float
        y-coordinate of center of area-sink
    R : float
        radius of area-sink
    N : float
        uniform infiltration rate of area-sink (positive for water going into
        the aquifer)
    layer : int
        layer where area-sink is applied (commonly layer 0, the top layer)
    label : string or None (default: None)
        label of the well

    """
    
    def __init__(self, model, xc=0, yc=0, R=1, N=0.001, layer=0, name='CircAreasink', label=None):
        self.storeinput(inspect.currentframe())
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layer, \
                         name=name, label=label)
        self.xc = xc
        self.yc = yc
        self.R = R
        self.N = N
        self.model.add_element(self)

    def __repr__(self):
        return self.name + ' at ' + str((self.xc, self.yc))

    def initialize(self):
        self.Rsq = self.R ** 2
        self.aq = self.model.aq.find_aquifer_data(self.xc, self.yc)
        self.aq.add_element(self)
        self.parameters = np.array([[self.N]])
        self.Rlarge = 500.0  # If R/lab > Rlarge, then we use asymptotic approximation to compute potential
        if self.aq.ilap:
            self.lab = self.aq.lab[1:]
            self.A = -self.aq.coef[self.layers, 1:] * self.R * self.lab
            self.B = self.aq.coef[self.layers, 1:] * self.R * self.lab
            self.C = self.aq.coef[self.layers, 1:] * self.lab ** 2
        else:
            self.lab = self.aq.lab
            self.A = -self.aq.coef[self.layers] * self.R * self.lab
            self.B = self.aq.coef[self.layers] * self.R * self.lab
            self.C = self.aq.coef[self.layers] * self.lab ** 2
        self.islarge = self.R / self.lab > self.Rlarge
        self.labsmall = self.lab[~self.islarge]
        self.labbig = self.lab[self.islarge]
        self.k1Roverlab = k1(self.R / self.labsmall)
        self.i1Roverlab = i1(self.R / self.labsmall)

    def potinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            r = np.sqrt((x - self.xc) ** 2 + (y - self.yc) ** 2)
            if r <= self.R:
                if aq.ilap:
                    rv[0, 0] = 0.25 * (self.Rsq - r ** 2)
                    rv[0, 1:] = self.A * self.K1RI0r(r) + self.C
                else:
                    rv[0] = self.A * self.K1RI0r(r) + self.C
            else:
                if aq.ilap:
                    rv[0, 0] = -0.5 * self.Rsq * np.log(r / self.R)
                    rv[0, 1:] = self.B * self.I1RK0r(r)
                else:
                    rv[0] = self.B * self.I1RK0r(r)
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            r = np.sqrt((x - self.xc) ** 2 + (y - self.yc) ** 2)
            if r <= self.R:
                if r > 1e-12:  # otherwise zero
                    if aq.ilap:
                        rv[0, 0, 0] = (x - self.xc) / 2
                        rv[1, 0, 0] = (y - self.yc) / 2
                        K1RI1r = self.K1RI1r(r)
                        rv[0, 0, 1:] = -self.A * K1RI1r * (x - self.xc) / (r * self.lab)
                        rv[1, 0, 1:] = -self.A * K1RI1r * (y - self.yc) / (r * self.lab)
                    else:
                        K1RI1r = self.K1RI1r(r) / self.lab
                        rv[0, 0] = -self.A * K1RI1r * (x - self.xc) / r
                        rv[1, 0] = -self.A * K1RI1r * (y - self.yc) / r
            else:
                if aq.ilap:
                    rv[0, 0, 0] = 0.5 * self.Rsq * (x - self.xc) / r ** 2
                    rv[1, 0, 0] = 0.5 * self.Rsq * (y - self.yc) / r ** 2
                    I1RK1r = self.I1RK1r(r)
                    rv[0, 0, 1:] = self.B * I1RK1r * (x - self.xc) / (r * self.lab)
                    rv[1, 0, 1:] = self.B * I1RK1r * (y - self.yc) / (r * self.lab)
                else:
                    I1RK1r = self.I1RK1r(r)
                    rv[0, 0] = self.B * I1RK1r * (x - self.xc) / (r * self.lab)
                    rv[1, 0] = self.B * I1RK1r * (y - self.yc) / (r * self.lab)
        else:
            r = np.sqrt((x - self.xc) ** 2 + (y - self.yc) ** 2)
            if r <= self.R:
                raise 'CircAreaSink should add constant inside inhomogeneity'
        return rv

    def K1RI0r(self, rin):
        rv = np.zeros(len(self.lab))
        if self.islarge.any():
            index = (self.R - rin) / self.labbig < 10
            if index.any():
                r = rin / self.labbig[index]
                R = self.R / self.labbig[index]
                rv[self.islarge * index] = np.sqrt(1 / (4 * r * R)) * np.exp(r - R) * \
                                   (1 + 3 / (8 * R) - 15 / (128 * R ** 2) + 315 / (3072 * R ** 3)) * \
                                   (1 + 1 / (8 * r) +  9 / (128 * r ** 2) + 225 / (3072 * r ** 3))
        if ~self.islarge.any():
            index = (self.R - rin) / self.labsmall < 10
            if index.any():
                r = rin / self.labsmall[index]
                rv[~self.islarge * index] = self.k1Roverlab[index] * i0(r)
        return rv

    def K1RI1r(self, rin):
        rv = np.zeros(len(self.lab))
        if self.islarge.any():
            index = (self.R - rin) / self.labbig < 10
            if index.any():
                r = rin / self.labbig[index]
                R = self.R / self.labbig[index]
                rv[self.islarge * index] = np.sqrt(1 / (4 * r * R)) * np.exp(r - R) * \
                                   (1 + 3 / (8 * R) - 15 / (128 * R ** 2) + 315 / (3072 * R ** 3)) * \
                                   (1 - 3 / (8 * r) - 15 / (128 * r ** 2) - 315 / (3072 * r ** 3))
        if ~self.islarge.any():
            index = (self.R - rin) / self.labsmall < 10
            if index.any():
                r = rin / self.labsmall[index]
                rv[~self.islarge * index] = self.k1Roverlab[index] * i1(r)
        return rv

    def I1RK0r(self, rin):
        rv = np.zeros(len(self.lab))
        if self.islarge.any():
            index = (rin - self.R) / self.labbig < 10
            if index.any():
                r = rin / self.labbig[index]
                R = self.R / self.labbig[index]
                rv[self.islarge * index] = np.sqrt(1 / (4 * r * R)) * np.exp(R - r) * \
                                   (1 - 3 / (8 * R) - 15 / (128 * R ** 2) - 315 / (3072 * R ** 3)) * \
                                   (1 - 1 / (8 * r) +  9 / (128 * r ** 2) - 225 / (3072 * r ** 3))
        if ~self.islarge.any():
            index = (self.R - rin) / self.labsmall < 10
            if index.any():
                r = rin / self.labsmall[index]
                rv[~self.islarge * index] = self.i1Roverlab[index] * k0(r)
        return rv

    def I1RK1r(self, rin):
        rv = np.zeros(len(self.lab))
        if self.islarge.any():
            index = (rin - self.R) / self.labbig < 10
            if index.any():
                r = rin / self.labbig[index]
                R = self.R / self.labbig[index]
                rv[self.islarge * index] = np.sqrt(1 / (4 * r * R)) * np.exp(R - r) * \
                                   (1 - 3 / (8 * R) - 15 / (128 * R ** 2) - 315 / (3072 * R ** 3)) * \
                                   (1 + 3 / (8 * r) - 15 / (128 * r ** 2) + 315 / (3072 * r ** 3))
        if ~self.islarge.any():
            index = (self.R - rin) / self.labsmall < 10
            if index.any():
                r = rin / self.labsmall[index]
                rv[~self.islarge * index] = self.i1Roverlab[index] * k1(r)
        return rv

    def qztop(self, x, y):
        rv = 0.0
        if np.sqrt((x - self.xc) ** 2 + (y - self.yc) ** 2) <= self.R:
            rv = -self.parameters[0, 0]  # minus cause the parameter is the infiltration rate
        return rv

    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer, direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        eps = 1e-8
        r1sq = (xyzt1[0] - self.xc) ** 2 + (xyzt1[1] - self.yc) ** 2
        r2sq = (xyzt2[0] - self.xc) ** 2 + (xyzt2[1] - self.yc) ** 2
        if (r1sq < self.Rsq and r2sq > self.Rsq ) or (r1sq > self.Rsq and r2sq < self.Rsq):
            changed = True
            x1, y1 = xyzt1[0:2]
            x2, y2 = xyzt2[0:2]
            a = (x2 - x1) ** 2 + (y2 - y1) ** 2
            b = 2 * ((x2 - x1) * (x1 - self.xc) + (y2 - y1) * (y1 - self.yc))
            c = self.xc ** 2 + self.yc ** 2 + x1 ** 2 + y1 ** 2 - 2 * (self.xc * x1 +self.yc * y1) - self.Rsq
            u1 = (-b - np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            u2 = (-b + np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
            if u1 > 0:
                u = u1 * (1.0 + eps) # Go just beyond circle
            else:
                u = u2 * (1.0 + eps) # Go just beyond circle
            xn = x1 + u * (x2 - x1)
            yn = y1 + u * (y2 - y1)
            zn = xyzt1[2] + u * (xyzt2[2] - xyzt1[2])
            xyztnew = xyzt1 + u * (xyzt2 - xyzt1)
        return changed, terminate, xyztnew, message



import numpy as np

def controlpoints(Ncp, z1, z2, eps=0, include_ends=False):
    #thetacp = np.arange(np.pi, 0, -np.pi/self.Ncp) - 0.5 * np.pi/self.Ncp
    # The following works MUCH better for a uniform head along the line
    thetacp = np.linspace(np.pi, 0, Ncp+2)[1:-1]
    if include_ends:
        Zcp = np.zeros(Ncp+2, 'D')
        Zcp[0] = -1
        Zcp[-1] = 1
        Zcp[1:-1] = np.cos(thetacp)
    else:
        #thetacp = np.arange(np.pi, 0, -np.pi/Ncp) - 0.5 * np.pi/Ncp
        Zcp = np.zeros(Ncp, 'D')
        Zcp.real = np.cos(thetacp)
    Zcp.imag = eps  # control point just on positive side (this is handy later on)
    zcp = Zcp * (z2 - z1) / 2.0 + 0.5 * (z1 + z2)
    return zcp.real, zcp.imag

def strengthinf_controlpoints(Ncp, Nlayers):
    # include_ends is False in comparison to function above
    thetacp = np.linspace(np.pi, 0, Ncp+2)[1:-1]
    Xcp = np.cos(thetacp)
    s = np.zeros((Ncp, Ncp))
    for i in range(Ncp):
        s[i] = Xcp[i] ** np.arange(Ncp)
    rv = np.zeros((Ncp * Nlayers, Ncp * Nlayers))
    for i in range(Ncp):
        for j in range(Nlayers):
            rv[i * Nlayers + j, j::Nlayers] = s[i]
    return rv


import inspect  # Used for storing the input

import matplotlib.pyplot as plt
import numpy as np
from scipy.special import k0, k1

from .element import Element
from .equation import MscreenWellEquation, PotentialEquation
from .trace import timtracelines

__all__ = ["WellBase", "Well", "HeadWell"]


class WellBase(Element):
    def __init__(self, model, xw=0, yw=0, Qw=100.0, rw=0.1, res=0.0, layers=0,
                 name="WellBase", label=None, xc=None, yc=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers,
                         name=name, label=label)
        # Defined here and not in Element as other elements can have multiple
        # parameters per layers
        self.nparam = len(self.layers)
        self.xw = float(xw)
        self.yw = float(yw)
        self.Qw = np.atleast_1d(Qw)
        self.rw = float(rw)
        self.res = float(res)
        self.xc = xc
        self.yc = yc
        self.model.add_element(self)

    def __repr__(self):
        return self.name + " at " + str((self.xw, self.yw))

    def initialize(self):
        if self.xc is None:
            self.xc = np.array([self.xw + self.rw])
        else:
            self.xc = np.atleast_1d(self.xc)
        if self.yc is None:
            self.yc = np.array([self.yw])
        else:
            self.yc = np.atleast_1d(self.yc)
        self.ncp = 1
        self.aq = self.model.aq.find_aquifer_data(self.xw, self.yw)
        self.aq.add_element(self)
        self.parameters = np.empty((self.nparam, 1))
        self.parameters[:, 0] = self.Qw
        self.resfac = self.res / (2 * np.pi * self.rw * self.aq.Haq[self.layers])

    def potinf(self, x, y, aq=None):
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            pot = np.zeros(aq.naq)
            r = np.sqrt((x - self.xw) ** 2 + (y - self.yw) ** 2)
            if r < self.rw:
                r = self.rw  # If at well, set to at radius
            if aq.ilap:
                pot[0] = np.log(r / self.rw) / (2 * np.pi)
                pot[1:] = -k0(r / aq.lab[1:]) / (2 * np.pi)
            else:
                pot[:] = -k0(r / aq.lab) / (2 * np.pi)
            rv[:] = self.aq.coef[self.layers] * pot
        return rv

    def disvecinf(self, x, y, aq=None):
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            qx = np.zeros(aq.naq)
            qy = np.zeros(aq.naq)
            rsq = (x - self.xw) ** 2 + (y - self.yw) ** 2
            r = np.sqrt(rsq)
            xminxw = x - self.xw
            yminyw = y - self.yw
            if r < self.rw:
                r = self.rw
                rsq = r ** 2
                xminxw = self.rw
                yminyw = 0.0
            if aq.ilap:
                qx[0] = -1 / (2 * np.pi) * xminxw / rsq
                qy[0] = -1 / (2 * np.pi) * yminyw / rsq
                kone = k1(r / aq.lab[1:])
                qx[1:] = -kone * xminxw / (r * aq.lab[1:]) / (2 * np.pi)
                qy[1:] = -kone * yminyw / (r * aq.lab[1:]) / (2 * np.pi)
            else:
                kone = k1(r / aq.lab)
                qx[:] = -kone * xminxw / (r * aq.lab) / (2 * np.pi)
                qy[:] = -kone * yminyw / (r * aq.lab) / (2 * np.pi)
            rv[0] = self.aq.coef[self.layers] * qx
            rv[1] = self.aq.coef[self.layers] * qy
        return rv

    def headinside(self):
        """The head inside the well

        Returns
        -------
        array (length number of screens)
            Head inside the well for each screen

        """

        h = self.model.head(self.xw + self.rw, self.yw, layers=self.layers)
        return h - self.resfac * self.parameters[:, 0]

    def discharge(self):
        """The discharge in each layer

        Returns
        -------
        array (length number of layers)
            Discharge in each screen with zeros for layers that are not
            screened

        """

        Q = np.zeros(self.aq.naq)
        Q[self.layers] = self.parameters[:, 0]
        return Q

    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer,
                    direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        if np.sqrt((xyzt2[0] - self.xw) ** 2 + (xyzt2[1] - self.yw) ** 2) < (
            hstepmax + self.rw
        ):
            if ltype == "a":
                if (layer == self.layers).any():  # in layer where well is screened
                    if (self.discharge()[layer] > 0 and direction > 0) or (
                        self.discharge()[layer] < 0 and direction < 0):
                        vx, vy, vz = self.model.velocity(*xyzt1[:-1])
                        tstep = np.sqrt(
                            (xyzt1[0] - self.xw) ** 2 + (xyzt1[1] - self.yw) ** 2
                        ) / np.sqrt(vx ** 2 + vy ** 2)
                        xnew = self.xw
                        ynew = self.yw
                        znew = xyzt1[2] + tstep * vz * direction
                        tnew = xyzt1[3] + tstep
                        xyztnew = np.array([xnew, ynew, znew, tnew])
                        changed = True
                        terminate = True
        if terminate:
            message = "reached element of type well"
            if self.label:
                message += " ({lab})".format(lab=self.label)
        return changed, terminate, [xyztnew], message

    def capzone(
        self, nt=10, zstart=None, hstepmax=10, vstepfrac=0.2, tmax=None,
        nstepmax=100, silent=".", *, metadata=False
    ):
        """Compute a capture zone

        Parameters
        ----------
        nt : int
            number of path lines
        zstart : scalar or None
            starting elevation of the path lines, middle of aquifer if None
        hstepmax : scalar
            maximum step in horizontal space
        vstepfrac : float
            maximum fraction of aquifer layer thickness during one step
        tmax : scalar
            maximum time
        nstepmax : scalar(int)
            maximum number of steps
        silent : boolean or string
            True (no messages), False (all messages), or '.'
            (print dot for each path line)

        Returns
        -------
        xyzt : list of arrays of x, y, z, and t values

        """
        xstart, ystart, zstart = self.capzonestart(nt, zstart)
        xyzt = timtracelines(self.model, xstart, ystart, zstart, -np.abs(hstepmax),
            vstepfrac=vstepfrac, tmax=tmax, nstepmax=nstepmax, silent=silent,
            metadata=metadata)
        return xyzt

    def capzonestart(self, nt, zstart):
        eps = 1e-1
        angle = np.arange(eps, 2 * np.pi, 2 * np.pi / nt)
        xstart = self.xw + (1 + eps) * self.rw * np.cos(angle)
        ystart = self.yw + (1 + eps) * self.rw * np.sin(angle)
        if zstart is None:
            zstart = self.aq.zaqbot[self.layers[0]] + 0.5 * self.aq.Haq[self.layers[0]]
        zstart = zstart * np.ones(nt)
        return xstart, ystart, zstart

    def plot(self, layer=None):
        if (layer is None) or (layer in self.layers):
            plt.plot(self.xw, self.yw, "k.")

    def plotcapzone(self, nt=10, zstart=None, hstepmax=20, vstepfrac=0.2,
                    tmax=365, nstepmax=100, silent=".", color=None,
                    orientation="hor", win=[-1e30, 1e30, -1e30, 1e30],
                    newfig=False, figsize=None, *, return_traces=False,
                    metadata=False):
        """Plot a capture zone

        Parameters
        ----------
        nt : int
            number of path lines
        zstart : scalar
            starting elevation of the path lines
        hstepmax : scalar
            maximum step in horizontal space
        vstepfrac : float
            maximum fraction of aquifer layer thickness during one step
        tmax : scalar
            maximum time
        nstepmax : scalar(int)
            maximum number of steps
        silent : boolean or string
            True (no messages), False (all messages), or '.'
            (print dot for each path line)
        color : color
        orientation : string
            'hor' for horizontal, 'ver' for vertical, or 'both' for both
        win : array_like (length 4)
            [xmin, xmax, ymin, ymax]
        newfig : boolean (default False)
            boolean indicating if new figure should be created
        figsize : tuple of integers, optional, default: None
            width, height in inches.

        """
        if not return_traces:
            metadata = True  # suppress future warning from timtraceline
        xstart, ystart, zstart = self.capzonestart(nt, zstart)
        traces = self.model.tracelines(xstart, ystart, zstart,
            hstepmax=-abs(hstepmax), vstepfrac=vstepfrac, tmax=tmax,
            nstepmax=nstepmax, silent=silent, color=color,
            orientation=orientation, win=win, newfig=newfig,
            figsize=figsize, return_traces=return_traces, metadata=metadata)
        if return_traces:
            return traces


class Well(WellBase, MscreenWellEquation):
    """
    Well Class to create a well with a specified discharge. The well
    may be screened in multiple layers. The resistance of the screen may
    be specified. The head is computed such that the discharge :math:`Q_i`
    in layer :math:`i` is computed as

    .. math::
        Q_i = 2\pi r_w(h_i - h_w)/c

    where :math:`c` is the resistance of the well screen and :math:`h_w` is
    the head inside the well. The total discharge is distributed over the
    screens such that :math:`h_w` is the same in each screened layer.

    Parameters
    ----------
    model : Model object
        model to which the element is added
    xw : float
        x-coordinate of the well
    yw : float
        y-coordinate of the well
    Qw : float
        total discharge of the well
    rw : float
        radius of the well
    res : float
        resistance of the well screen
    layers : int, array or list
        layer (int) or layers (list or array) where well is screened
    label : string or None (default: None)
        label of the well
    xc : float
        x-location of control point (default None, which puts it at xw)
    yc : float
        y-location of control point (default None, which puts it at yw + rw)

    Examples
    --------
    >>> ml = Model3D(kaq=10, z=np.arange(20, -1, -2), kzoverkh=0.1)
    >>> Well(ml, 100, 200, 1000, layers=[0, 1, 2, 3])

    """

    def __init__(self, model, xw=0, yw=0, Qw=100.0, rw=0.1, \
                 res=0.0, layers=0, label=None, xc=None, yc=None):
        self.storeinput(inspect.currentframe())
        WellBase.__init__(self, model, xw, yw, Qw, rw, res, \
                          layers=layers, name="Well", label=label, xc=xc, yc=yc)
        self.Qc = float(Qw)
        if self.nlayers == 1:
            self.nunknowns = 0
        else:
            self.nunknowns = self.nparam

    def initialize(self):
        WellBase.initialize(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol


class HeadWell(WellBase, PotentialEquation):
    """
    HeadWell Class to create a well with a specified head inside the well.
    The well may be screened in multiple layers. The resistance of the screen
    may be specified. The head is computed such that the discharge :math:`Q_i`
    in layer :math:`i` is computed as

    .. math::
        Q_i = 2\pi r_w(h_i - h_w)/c

    where :math:`c` is the resistance of the well screen and :math:`h_w` is
    the head inside the well.

    Parameters
    ----------
    model : Model object
        model to which the element is added
    xw : float
        x-coordinate of the well
    yw : float
        y-coordinate of the well
    hw : float
        head inside the well
    rw : float
        radius of the well
    res : float
        resistance of the well screen
    layers : int, array or list
        layer (int) or layers (list or array) where well is screened
    label : string (default: None)
        label of the well
    xc : float
        x-location of control point (default None, which puts it at xw)
    yc : float
        y-location of control point (default None, which puts it at yw + rw)

    """

    def __init__(self, model, xw=0, yw=0, hw=10, rw=0.1, res=0, layers=0,
                 label=None, xc=None, yc=None):
        self.storeinput(inspect.currentframe())
        WellBase.__init__(self, model, xw, yw, 0.0, rw, res, layers=layers,
                          name="HeadWell", label=label, xc=xc, yc=yc)
        self.hc = hw
        self.nunknowns = self.nparam

    def initialize(self):
        WellBase.initialize(self)
        self.pc = self.hc * self.aq.T[self.layers]  # Needed in solving

    def setparams(self, sol):
        self.parameters[:, 0] = sol



import numpy as np
from .linesink import LineSinkHoBase
from .equation import HeadDiffEquation2, DisvecDiffEquation2, IntDisVecEquation
from .controlpoints import controlpoints
# needed for testing
# from .equation import DisvecDiffEquation


class IntHeadDiffLineSink(LineSinkHoBase, HeadDiffEquation2):
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0,
                 order=0, ndeg=3, layers=None, label=None, addtomodel=True,
                 aq=None, aqin=None, aqout=None):
        if layers is None:
            layers = np.arange(model.aq.naq)
        LineSinkHoBase.__init__(self, model, x1, y1, x2, y2, Qls=0, \
                                layers=layers, order=order, \
                                name='IntHeadDiffLineSink', label=label, \
                                addtomodel=addtomodel, aq=aq)
        self.inhomelement = True
        self.ndeg = ndeg
        self.Xleg, self.wleg = np.polynomial.legendre.leggauss(self.ndeg)
        self.nunknowns = self.nparam
        self.aqin = aqin
        self.aqout = aqout

    def initialize(self):
        LineSinkHoBase.initialize(self)
        self.xcin, self.ycin = controlpoints(self.ncp - 1, self.z1, self.z2,
                                             eps=1e-6, include_ends=True)
        self.xcout, self.ycout = controlpoints(self.ncp - 1, self.z1, self.z2,
                                               eps=-1e-6, include_ends=True)
        if self.aqin is None:
            self.aqin = self.model.aq.find_aquifer_data(self.xcin[1],
                                                        self.ycin[1])
        if self.aqout is None:
            self.aqout = self.model.aq.find_aquifer_data(self.xcout[1],
                                                         self.ycout[1])

    def setparams(self, sol):
        self.parameters[:, 0] = sol

    #def changetrace(self, xyzt1, xyzt2, layer, ltype):
    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer, direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        if (ltype == 'a'):
            eps = 1e-8
            za = xyzt1[0] + xyzt1[1] * 1j
            zb = xyzt2[0] + xyzt2[1] * 1j
            Za = (2 * za - (self.z1 + self.z2)) / (self.z2 - self.z1)
            Zb = (2 * zb - (self.z1 + self.z2)) / (self.z2 - self.z1)
            if Za.imag * Zb.imag < 0:
                Xa, Ya = Za.real, Za.imag
                Xb, Yb = Zb.real, Zb.imag
                X = Xa - Ya * (Xb - Xa) / (Yb - Ya)
                if abs(X) <= 1:
                    Z = X + eps * np.sign(Yb) * 1j  # steps to side of Yb
                    z = 0.5 * ((self.z2 - self.z1) * Z + self.z1 + self.z2)
                    xnew, ynew = z.real, z.imag
                    dold = abs(za - zb)
                    dnew = np.sqrt((xnew - xyzt1[0]) ** 2 + (ynew - xyzt1[1]) ** 2)
                    znew = xyzt1[2] + dnew / dold * (xyzt2[2] - xyzt1[2])
                    tnew = xyzt1[3] + dnew / dold * (xyzt2[3] - xyzt1[3])
                    xyztnew = np.array([xnew, ynew, znew, tnew])
                    changed = True
        return changed, terminate, [xyztnew], message


class IntFluxDiffLineSink(LineSinkHoBase, DisvecDiffEquation2):
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, layers=None,
                 order=0, ndeg=3, label=None, addtomodel=True, aq=None,
                 aqin=None, aqout=None):
        if layers is None:
            layers = list(range(model.aq.naq))
        LineSinkHoBase.__init__(self, model, x1, y1, x2, y2, Qls=0,
                                layers=layers, order=order,
                                name='IntFluxDiffLineSink', label=label,
                                addtomodel=addtomodel, aq=aq)
        self.inhomelement = True
        self.ndeg = ndeg
        self.Xleg, self.wleg = np.polynomial.legendre.leggauss(self.ndeg)
        self.nunknowns = self.nparam
        self.aqin = aqin
        self.aqout = aqout

    def initialize(self):
        LineSinkHoBase.initialize(self)
        self.xcin, self.ycin = controlpoints(self.ncp - 1, self.z1, self.z2,
                                             eps=1e-6, include_ends=True)
        self.xcout, self.ycout = controlpoints(self.ncp - 1, self.z1, self.z2,
                                               eps=-1e-6, include_ends=True)
        if self.aqin is None:
            self.aqin = self.model.aq.find_aquifer_data(self.xcin[0],
                                                        self.ycin[0])
        if self.aqout is None:
            self.aqout = self.model.aq.find_aquifer_data(self.xcout[0],
                                                         self.ycout[0])

    def setparams(self, sol):
        self.parameters[:, 0] = sol

    #def changetrace(self, xyzt1, xyzt2, layer, ltype):
    def changetrace(self, xyzt1, xyzt2, aq, layer, ltype, modellayer, direction, hstepmax):
        changed = False
        terminate = False
        xyztnew = 0
        message = None
        if (ltype == 'a'):
            eps = 1e-8
            za = xyzt1[0] + xyzt1[1] * 1j
            zb = xyzt2[0] + xyzt2[1] * 1j
            Za = (2 * za - (self.z1 + self.z2)) / (self.z2 - self.z1)
            Zb = (2 * zb - (self.z1 + self.z2)) / (self.z2 - self.z1)
            if Za.imag * Zb.imag < 0:
                Xa, Ya = Za.real, Za.imag
                Xb, Yb = Zb.real, Zb.imag
                X = Xa - Ya * (Xb - Xa) / (Yb - Ya)
                if abs(X) <= 1:
                    Z = X + eps * np.sign(Yb) * 1j  # steps to side of Yb
                    z = 0.5 * ((self.z2 - self.z1) * Z + self.z1 + self.z2)
                    xnew, ynew = z.real, z.imag
                    dold = abs(za - zb)
                    dnew = np.sqrt((xnew - xyzt1[0]) ** 2 + (ynew - xyzt1[1]) ** 2)
                    znew = xyzt1[2] + dnew / dold * (xyzt2[2] - xyzt1[2])
                    tnew = xyzt1[3] + dnew / dold * (xyzt2[3] - xyzt1[3])
                    xyztnew = np.array([xnew, ynew, znew, tnew])
                    changed = True
                    #return True, False, xyztnew
        return changed, terminate, [xyztnew], message


class IntFluxLineSink(LineSinkHoBase, IntDisVecEquation):
    """Element to set numerically integrated flux
    along linesink to 0. Used in BuildingPit element
    """

    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, layers=None,
                 order=0, ndeg=3, label=None, addtomodel=True, aq=None,
                 aqin=None, aqout=None):
        if layers is None:
            layers = list(range(model.aq.naq))
        LineSinkHoBase.__init__(self, model, x1, y1, x2, y2, Qls=0,
                                layers=layers, order=order,
                                name='IntFluxDiffLineSink', label=label,
                                addtomodel=addtomodel, aq=aq)
        self.inhomelement = True
        self.ndeg = ndeg
        self.Xleg, self.wleg = np.polynomial.legendre.leggauss(self.ndeg)
        self.nunknowns = self.nparam
        self.aqin = aqin
        self.aqout = aqout

    def initialize(self):
        LineSinkHoBase.initialize(self)

        # recalculated with ncp - 1 points instead of ncp points
        self.xcin, self.ycin = controlpoints(self.ncp - 1, self.z1, self.z2,
                                             eps=1e-6, include_ends=True)
        self.xcout, self.ycout = controlpoints(self.ncp - 1, self.z1, self.z2,
                                               eps=-1e-6, include_ends=True)

        if self.aqin is None:
            self.aqin = self.model.aq.find_aquifer_data(self.xcin[0],
                                                        self.ycin[0])
        if self.aqout is None:
            self.aqout = self.model.aq.find_aquifer_data(self.xcout[0],
                                                         self.ycout[0])
        # set control points x,y depending on which side of
        # the boundary the element is for
        if self.aq == self.aqin:
            self.xc = self.xcin
            self.yc = self.ycin
        if self.aq == self.aqout:
            self.xc = self.xcout
            self.yc = self.ycout

    def setparams(self, sol):
        self.parameters[:, 0] = sol



import numba
import numpy as np

__all__ = [
    "potbesldho",
    "potbeslsho",
    "potbesldv",
    "potbeslsv",
    "disbesldho",
    "disbeslsho",
    "disbesldv",
    "disbeslsv",
]


def initialize():
    pass


@numba.njit(nogil=True)
def make_rbinom():
    rrange = np.arange(9.0)
    rbinom = np.zeros((9, 9))
    for n in range(9):
        for m in range(9):
            if m > n:
                rbinom[n, m] = 1
            else:
                rbinom[n, m] = np.prod(rrange[m + 1 : n + 1]) / np.prod(
                    rrange[1 : n - m + 1]
                )
    return rbinom


RBINOM = make_rbinom()
AC = np.zeros(9)
BC = np.zeros(9)
AC1 = np.zeros(9)
BC1 = np.zeros(9)
NTERMS = 8

# Coefficients of Table 1
AC[0] = -0.500004211065677e0
BC[0] = 0.115956920789028e0
AC[1] = -0.124989431448700e0
BC[1] = 0.278919134951974e0
AC[2] = -0.781685695640975e-2
BC[2] = 0.252752008621167e-1
AC[3] = -0.216324415010288e-3
BC[3] = 0.841879407543506e-3
AC[4] = -0.344525393452639e-5
BC[4] = 0.152425102734818e-4
AC[5] = -0.315133836828774e-7
BC[5] = 0.148292488399579e-6
AC[6] = -0.296636186265427e-9
BC[6] = 0.157622547107156e-8
AC[7] = -0.313689942474032e-12
BC[7] = 0.117975437124933e-11
AC[8] = -0.112031912249579e-13
BC[8] = 0.655345107753534e-13

# Coefficients of K1
AC1[0] = 0.250000197208863e0
BC1[0] = -0.307966963840932e0
AC1[1] = 0.312495047266388e-1
BC1[1] = -0.853676915840295e-1
AC1[2] = 0.130228768540005e-2
BC1[2] = -0.464343185899275e-2
AC1[3] = 0.270943632576982e-4
BC1[3] = -0.112338260301993e-3
AC1[4] = 0.341642640876988e-6
BC1[4] = -0.157491472277549e-5
AC1[5] = 0.271286480571077e-8
BC1[5] = -0.133414321160211e-7
AC1[6] = 0.197096143802491e-10
BC1[6] = -0.106342159633141e-9
AC1[7] = 0.329351103353054e-13
BC1[7] = -0.159531523882074e-12
AC1[8] = 0.573031034976631e-15
BC1[8] = -0.340195779923156e-14


@numba.njit(nogil=True)
def prepare_z(x, y, z1, z2):
    zin = np.complex(x, y)
    z1in = z1
    z2in = z2
    Lin = abs(z2in - z1in)
    z = (2.0 * zin - (z1in + z2in)) / (z2in - z1in)
    zplus1 = z + 1.0
    zmin1 = z - 1.0
    # If at cornerpoint, move slightly
    if abs(zplus1) < 1.0e-8 * 2.0 / Lin:
        zplus1 = zplus1 + 1.0e-8
    if abs(zmin1) < 1.0e-8 * 2.0 / Lin:
        zmin1 = zmin1 + 1.0e-8
    return zin, z1in, z2in, Lin, z, zplus1, zmin1


@numba.njit(nogil=True)
def potbeslsho(x, y, z1, z2, labda, order, ilap, naq):
    """
    Parameters
    ---------- 
       x,y: Point where potential is computed
       z1: Complex begin point of line-sink
       z2: Complex end point of line-sink
       labda(naq): labda's (zero for first labda if Laplace)
       order: Order of the line-sink
       ilap: equals 1 when first value is Laplace line-sink and first labda equals zero
       naq: Number of aquifers
       rv(naq): Array to store return value (must be pre-allocated)
    Returns
    --------
       rv(naq): Potentials. First spot is Laplace value if ilap=1
    """

    rv = np.zeros(naq)

    # lstype = 1 means line-sink
    lstype = 1

    # Radius of convergence
    Rconv = 7.0

    # if (ilap==1) :
    #    istart = 1
    # else:
    #    istart = 0
    zin, z1in, z2in, Lin, z, zplus1, zmin1 = prepare_z(x, y, z1, z2)
    # Laplace linesink
    if ilap == 1:
        power = order + 1
        pcor = np.complex(0.0, 0.0)
        for n in range(1, int((power + 1) / 2) + 1):
            pcor = pcor + z ** (power - 2 * n + 1) / (2 * n - 1)
        pcor = 2.0 * pcor
        comega = (
            z ** power * np.log((zmin1) / (zplus1))
            + pcor
            - np.log(zmin1)
            + (-1.0) ** power * np.log(zplus1)
        )
        comega = -comega * Lin / (4.0 * np.pi * power)
        rv[0] = np.real(comega)
    # N-1 leakage factors
    for i in range(ilap, naq):
        # Check whether entire linesink is outside radius of convergence
        # Outside if |z-zc|>L/2+7lab, and thus |Z|>1+7lab*2/L, or |zeta|>1/biglab+7 (zeta is called z here)
        biglab = 2.0 * labda[i] / Lin
        z = (2.0 * zin - (z1in + z2in)) / (z2in - z1in) / biglab

        if abs(z) < (Rconv + 1.0 / biglab):
            pot = IntegralF(zin, z1in, z2in, Lin, labda[i], order, Rconv, lstype)
            rv[i] = -Lin / 2.0 * pot
        else:
            rv[i] = 0.0
    return rv


@numba.njit(nogil=True)
def potbeslsv(x, y, z1, z2, lab, order, ilap, naq):
    # Check if endpoints need to be adjusted using the largest labda (the first one)
    pot = np.zeros((order + 1, naq))
    for n in range(0, order + 1):
        pot[n, 0 : naq + 1] = potbeslsho(x, y, z1, z2, lab, n, ilap, naq)
    return pot


@numba.njit(nogil=True)
def disbeslsho(x, y, z1, z2, labda, order, ilap, naq):
    # Input:
    #   x,y: Point where discharge is computed
    #   z1: Complex begin point of line-sink
    #   z2: Complex end point of line-sink
    #   labdain(Naquifers): Array with zero in first spot and labda's in remaining spots
    #   order: Order of the line-sink
    #   Naquifers: Number of aquifers
    #   rvx(Naquifers),rvy(Naquifers): Arrays to store return values
    # Output:
    #   rvx(Naquifers),rvy(Naquifers): Values of Qx and Qy with Laplace value in first spot
    #                   and mod.Helmholtz potentials in remaining spots

    rv = np.zeros((2, naq))
    # Radius of convergence
    Rconv = 7.0

    # lstype = 1 means line-sink
    lstype = 1

    zin, z1in, z2in, Lin, z, zplus1, zmin1 = prepare_z(x, y, z1, z2)

    pcor = 0.0
    # Laplace linesink
    if ilap == 1:
        pcor = np.complex(0.0, 0.0)
        for n in range(1, int((order + 1) / 2) + 1):
            pcor = pcor + float(order - 2 * n + 2) * z ** (order + 1 - 2 * n) / float(
                2 * n - 1
            )

        pcor = 2.0 * pcor

        cdum = 1.0 / (
            order + 1
        )  # Without this intermediate statement it didn't seem to work
        wdis = float(order + 1) * z ** order * np.log((zmin1) / (zplus1)) + pcor

        wdis = (
            wdis
            + (z ** (order + 1) - 1.0) / zmin1
            - (z ** (order + 1) - (-1.0) ** (order + 1)) / zplus1
        )

        wdis = wdis * Lin / 2.0 / (z2in - z1in) / np.pi * cdum

        rv[0, 0] = np.real(wdis)
        rv[1, 0] = -np.imag(wdis)

    for i in range(ilap, naq):
        wdis = np.complex(0.0, 0.0)

        # Check whether entire linesink is outside radius of convergence
        # Outside if |z-zc|>L/2+7lab, and thus |Z|>1+7lab*2/L, or |zeta|>1/biglab+7 (zeta is called z here)
        biglab = 2.0 * labda[i] / Lin
        z = (2.0 * zin - (z1in + z2in)) / (z2in - z1in) / biglab

        if abs(z) < (Rconv + 1.0 / biglab):
            wdis = IntegralG(zin, z1in, z2in, Lin, labda[i], order, Rconv, lstype)
            wdis = 2.0 * Lin / (z2in - z1in) / biglab * wdis

            rv[0, i] = np.real(wdis)
            rv[1, i] = -np.imag(wdis)
        else:
            rv[0, i] = 0.0
            rv[1, i] = 0.0

    return rv


@numba.njit(nogil=True)
def disbeslsv(x, y, z1, z2, lab, order, ilap, naq):
    # locals
    qxqy = np.zeros((2 * (order + 1), naq))
    # Check if endpoints need to be adjusted using the largest labda (the first one)
    for n in range(0, order + 1):
        rv = disbeslsho(x, y, z1, z2, lab, n, ilap, naq)
        qxqy[n, 0 : naq + 1] = rv[0, 0 : naq + 1]
        qxqy[n + order + 1, 0 : naq + 1] = rv[1, 0 : naq + 1]
    return qxqy


@numba.njit(nogil=True)
def potbesldho(x, y, z1, z2, labda, order, ilap, naq):

    # Input:
    #   x,y: Point where potential is computed
    #   z1: Complex begin point of line-doublet
    #   z2: Complex end point of line-doublet
    #   labda(naq): labda's (zero for first labda if Laplace)
    #   order: Order of the line-doublet
    #   ilap: equals 1 when first value is Laplace line-doublet and first labda equals zero
    #   naq: Number of aquifers
    #   rv(naq): Array to store return value (must be pre-allocated)
    # Output:
    #   rv(naq): Potentials. First spot is Laplace value if ilap=1

    rv = np.zeros(naq)

    # Radius of convergence
    Rconv = 7.0

    # lstype=2 means line-doublet
    lstype = 2
    zin, z1in, z2in, Lin, z, zplus1, zmin1 = prepare_z(x, y, z1, z2)

    # Laplace line-doublet
    if ilap == 1:
        comega = z ** order * np.log(zmin1 / zplus1)
        qm = np.complex(0.0, 0.0)
        for n in range(1, int((order + 1) / 2) + 1):
            qm = qm + z ** (order - 2.0 * float(n) + 1.0) / (2.0 * float(n) - 1.0)

        comega = 1.0 / (2.0 * np.pi * np.complex(0.0, 1.0)) * (comega + 2.0 * qm)
        rv[0] = np.real(comega)

    # N-1 leakage factors
    for i in range(ilap, naq):
        pot = 0.0
        # Check whether entire linedoublet is outside radius of convergence
        # Outside if |z-zc|>L/2+7lab, and thus |Z|>1+7lab*2/L, or |zeta|>1/biglab+7 (zeta is called z here)
        biglab = 2.0 * labda[i] / Lin
        z = (2.0 * zin - (z1in + z2in)) / (z2in - z1in) / biglab

        if abs(z) < (Rconv + 1.0 / biglab):
            m1, m2, NLS = findm1m2(zin, z1in, z2in, Lin, labda[i], Rconv)
            comega = np.complex(0.0, 0.0)
            if m1 > 0:  # Otherwise outside radius of convergence
                z1new = z1in + float(m1 - 1) / float(NLS) * (z2in - z1in)
                z2new = z1in + float(m2) / float(NLS) * (z2in - z1in)
                del0 = float(1 - m1 - m2 + NLS) / float(1 - m1 + m2)
                ra = float(NLS) / float(1 + m2 - m1)
                comega = IntegralLapLineDipole(zin, z1new, z2new, del0, ra, order)

            pot = IntegralF(zin, z1in, z2in, Lin, labda[i], order, Rconv, lstype)
            rv[i] = (
                np.real(comega / np.complex(0.0, 1.0)) + np.imag(z) / biglab * pot
            )  # Note that z is really zeta in analysis
        else:
            rv[i] = 0.0

    return rv


@numba.njit(nogil=True)
def potbesldv(x, y, z1, z2, lab, order, ilap, naq):
    pot = np.zeros((order + 1, naq))
    # Check if endpoints need to be adjusted using the largest labda (the first one)
    for n in range(0, order + 1):
        pot[n, 0 : naq + 1] = potbesldho(x, y, z1, z2, lab, n, ilap, naq)
    return pot


@numba.njit(nogil=True)
def disbesldho(x, y, z1, z2, labda, order, ilap, naq):
    # Input:
    #   x,y: Point where discharge is computed
    #   z1: Complex begin point of line-sink
    #   z2: Complex end point of line-sink
    #   labdain(Naquifers): Array with zero in first spot and labda's in remaining spots
    #   order: Order of the line-sink
    #   naq: Number of aquifers
    # Output:
    #   rv(2, Naquifers),rvy(Naquifers): Values of Qx and Qy with Laplace value in first spot
    #                   and mod.Helmholtz potentials in remaining spots
    rv = np.zeros((2, naq))
    # Radius of convergence
    Rconv = 7.0

    # lstype=2 means line-doublet
    lstype = 2

    zin, z1in, z2in, Lin, z, zplus1, zmin1 = prepare_z(x, y, z1, z2)

    # Laplace line-doublet
    qm = np.complex(0.0, 0.0)
    if ilap == 1:
        if order == 0:
            wdis = -(1.0 / zmin1 - 1.0 / zplus1) / (
                np.pi * np.complex(0.0, 1.0) * (z2in - z1in)
            )
        else:
            wdis = float(order) * z ** (order - 1) * np.log(zmin1 / zplus1)
            wdis = wdis + z ** order * (1.0 / zmin1 - 1.0 / zplus1)
            qm = np.complex(0.0, 0.0)
            if order > 1:  # To avoid a possible problem of 0 * 0^(-1)
                for n in range(1, int(order / 2) + 1):
                    qm = qm + float(order - 2 * n + 1) * z ** (order - 2 * n) / float(
                        2 * n - 1
                    )

            wdis = -(wdis + 2.0 * qm) / (np.pi * np.complex(0.0, 1.0) * (z2in - z1in))

        rv[0, 0] = np.real(wdis)
        rv[1, 0] = -np.imag(wdis)

    # N-1 or N leakage factors
    for i in range(ilap, naq):
        wdis = np.complex(0.0, 0.0)

        # Check whether entire line-doublet is outside radius of convergence
        # Outside if |z-zc|>L/2+7lab, and thus |Z|>1+7lab*2/L, or |zeta|>1/biglab+7 (zeta is called z here)
        biglab = 2.0 * labda[i] / Lin
        z = (2.0 * zin - (z1in + z2in)) / (z2in - z1in) / biglab

        if abs(z) < (Rconv + 1.0 / biglab):
            m1, m2, NLS = findm1m2(zin, z1in, z2in, Lin, labda[i], Rconv)
            wdis1 = np.complex(0.0, 0.0)
            if m1 > 0:
                z1new = z1in + float(m1 - 1) / float(NLS) * (z2in - z1in)
                z2new = z1in + float(m2) / float(NLS) * (z2in - z1in)
                del0 = float(1 - m1 - m2 + NLS) / float(1 - m1 + m2)
                ra = float(NLS) / float(1 + m2 - m1)
                wdis1 = IntegralLapLineDipoleDis(zin, z1new, z2new, del0, ra, order)
                wdis1 = -2.0 * wdis1 / (np.complex(0.0, 1.0) * (z2new - z1new))

            pot = IntegralF(zin, z1in, z2in, Lin, labda[i], order, Rconv, lstype)
            wdis2 = IntegralG(zin, z1in, z2in, Lin, labda[i], order, Rconv, lstype)
            wdis3 = pot / (2.0 * np.complex(0.0, 1.0)) + wdis2 * np.imag(z)

            wdis = wdis1 - 4.0 * wdis3 / (biglab ** 2 * (z2in - z1in))
        rv[0, i] = np.real(wdis)
        rv[1, i] = -1.0 * np.imag(wdis)
    return rv


@numba.njit(nogil=True)
def disbesldv(x, y, z1, z2, lab, order, ilap, naq):
    qxqy = np.zeros((2 * (order + 1), naq))
    # Check if endpoints need to be adjusted using the largest labda (the first one)
    for n in range(0, order + 1):
        rv = disbesldho(x, y, z1, z2, lab, n, ilap, naq)
        qxqy[n, 0 : naq + 1] = rv[0, 0:naq]
        qxqy[n + order + 1, 0 : naq + 1] = rv[1, 0 : naq + 1]
    return qxqy


@numba.njit(nogil=True)
def IntegralF(zin, z1in, z2in, Lin, labda, order, Rconv, lstype):
    czmzbarp = np.full(NTERMS + 1, np.complex(0.0, 0.0))
    cgamma = np.full((NTERMS + 1, NTERMS + 1), np.complex(0.0, 0.0))
    calphat = np.full(2 * NTERMS + 1, np.complex(0.0, 0.0))
    cbetat = np.full(2 * NTERMS + 1, np.complex(0.0, 0.0))
    cc = np.full(order + 2, np.complex(0.0, 0.0))
    calpha = np.full(2 * NTERMS + order + 1, np.complex(0.0, 0.0))
    cbeta = np.full(2 * NTERMS + order + 1, np.complex(0.0, 0.0))

    m1, m2, NLS = findm1m2(zin, z1in, z2in, Lin, labda, Rconv)
    if m1 == 0:
        # pot = 0.0
        return 0.0

    # Compute zeta (called z here). This is the regular value of the entire element
    L = abs(z2in - z1in)
    biglab = 2.0 * labda / L
    z = (2.0 * zin - (z1in + z2in)) / (z2in - z1in) / biglab
    zbar = np.conj(z)

    # Coefficients gamma(n,m), Eq. 21
    # Store coefficents in matrix.
    for n in range(0, NTERMS + 1):
        czmzbarp[n] = (z - zbar) ** n

    for n in range(0, NTERMS + 1):
        for m in range(0, n + 1):
            cgamma[n, m] = RBINOM[n, m] * czmzbarp[n - m]

    # Eq. 23 These coefficients should be modified for a higher order linesink
    for n in range(0, 2 * NTERMS + 1):
        calphat[n] = np.complex(0.0, 0.0)
        cbetat[n] = np.complex(0.0, 0.0)
        for m in range(max(0, n - NTERMS), int(n / 2) + 1):
            if lstype == 1:
                calphat[n] = calphat[n] + AC[n - m] * cgamma[n - m, m]
                cbetat[n] = cbetat[n] + BC[n - m] * cgamma[n - m, m]
            else:
                calphat[n] = calphat[n] + AC1[n - m] * cgamma[n - m, m]
                cbetat[n] = cbetat[n] + BC1[n - m] * cgamma[n - m, m]

    # Compute coefficients of delta^p
    for m in range(0, order + 1):
        cc[m] = RBINOM[order, m] * z ** (order - m) * biglab ** order
    if order > 0:
        for n in range(0, 2 * NTERMS + order + 1):
            calpha[n] = np.complex(0.0, 0.0)
            cbeta[n] = np.complex(0.0, 0.0)
            for m in range(max(0, n - 2 * NTERMS), min(n, order) + 1):
                calpha[n] = calpha[n] + cc[m] * calphat[n - m]
                cbeta[n] = cbeta[n] + cc[m] * cbetat[n - m]
    else:
        calpha = calphat
        cbeta = cbetat

    # Evaluation of integral, Eq. 25
    cInt = np.complex(0.0, 0.0)
    del1 = -1.0 + 2.0 * (float(m1) - 1.0) / float(NLS)
    del2 = -1.0 + 2.0 * float(m2) / float(NLS)
    cd1minz = del1 / biglab - z
    cd2minz = del2 / biglab - z
    if abs(cd1minz) < 1.0e-8 / labda:
        cd1minz = cd1minz + 1.0e-8
    if abs(cd2minz) < 1.0e-8 / labda:
        cd2minz = cd2minz + 1.0e-8
    cln1 = np.log(cd1minz)
    cln2 = np.log(cd2minz)
    for n in range(0, 2 * NTERMS + order + 1):
        cInt = cInt + (
            2.0 * calpha[n] * cln2 - 2.0 * calpha[n] / (n + 1) + cbeta[n]
        ) * (cd2minz) ** (n + 1) / float(n + 1)
        cInt = cInt - (
            2.0 * calpha[n] * cln1 - 2.0 * calpha[n] / (n + 1) + cbeta[n]
        ) * (cd1minz) ** (n + 1) / float(n + 1)
    pot = np.real(cInt) * biglab / (2.0 * np.pi)
    return pot


@numba.njit(nogil=True)
def IntegralG(zin, z1in, z2in, Lin, labda, order, Rconv, lstype):
    czmzbarp = np.full(NTERMS + 1, np.complex(0.0, 0.0))
    cgamma = np.full((NTERMS + 1, NTERMS + 1), np.complex(0.0, 0.0))
    cahat = np.full(2 * (NTERMS - 1) + 1, np.complex(0.0, 0.0))
    cbhat = np.full(2 * (NTERMS - 1) + 1, np.complex(0.0, 0.0))
    calphat = np.full(2 * NTERMS + 1, np.complex(0.0, 0.0))
    cbetat = np.full(2 * NTERMS + 1, np.complex(0.0, 0.0))
    cc = np.full(order + 2, np.complex(0.0, 0.0))
    calpha = np.full(2 * NTERMS + order + 1, np.complex(0.0, 0.0))
    cbeta = np.full(2 * NTERMS + order + 1, np.complex(0.0, 0.0))

    biglabin = 2.0 * labda / Lin

    m1, m2, NLS = findm1m2(zin, z1in, z2in, Lin, labda, Rconv)
    if m1 == 0:
        # wdis = np.complex(0.0, 0.0)
        return np.complex(0.0, 0.0)

    # Compute zeta (called z here). This is the regular value of the entire element
    L = abs(z2in - z1in)
    biglab = 2.0 * labda / L
    z = (2.0 * zin - (z1in + z2in)) / (z2in - z1in) / biglab
    zbar = np.conj(z)

    # Coefficients gamma(n,m), Eq. 21
    # Store coefficents in matrix.
    for n in range(0, NTERMS + 1):
        czmzbarp[n] = (z - zbar) ** n

    for n in range(0, NTERMS + 1):
        for m in range(0, n + 1):
            cgamma[n, m] = RBINOM[n, m] * czmzbarp[n - m]

    # Integral g1
    # Implemented with different z, rather than Delta1 and Delta2
    z1 = z1in + float(m1 - 1) / float(NLS) * (z2in - z1in)
    z2 = z1in + float(m2) / float(NLS) * (z2in - z1in)
    del0 = float(1 - m1 - m2 + NLS) / float(1 - m1 + m2)
    ra = float(NLS) / float(1 + m2 - m1)
    # comega = np.complex(0.0, 0.0)
    comega = IntegralLapLineDipole(zin, z1, z2, del0, ra, order)

    if lstype == 1:
        g1 = -AC[0] * biglabin * comega
        # Integral g2
        # Compute hat coefficients
        for n in range(0, NTERMS):
            cahat[n] = float(n + 1) * AC[n + 1]
            cbhat[n] = AC[n + 1] + float(n + 1) * BC[n + 1]
    else:
        g1 = -AC1[0] * biglabin * comega
        # Integral g2
        # Compute hat coefficients
        for n in range(0, NTERMS):
            cahat[n] = float(n + 1) * AC1[n + 1]
            cbhat[n] = AC1[n + 1] + float(n + 1) * BC1[n + 1]

    # Eq. 23
    for n in range(0, 2 * NTERMS):
        calphat[n] = np.complex(0.0, 0.0)
        cbetat[n] = np.complex(0.0, 0.0)
        for m in range(max(0, n - NTERMS + 1), int((n + 1) / 2) + 1):
            calphat[n] = calphat[n] + cahat[n - m] * cgamma[n - m + 1, m]
            cbetat[n] = cbetat[n] + cbhat[n - m] * cgamma[n - m + 1, m]

    # Compute coefficients of delta^p
    for m in range(0, order + 1):
        cc[m] = RBINOM[order, m] * z ** (order - m) * biglab ** order
    if order > 0:
        for n in range(0, 2 * NTERMS + order):
            calpha[n] = np.complex(0.0, 0.0)
            cbeta[n] = np.complex(0.0, 0.0)
            for m in range(max(0, n - 2 * NTERMS + 1), min(n, order) + 1):
                calpha[n] = calpha[n] + cc[m] * calphat[n - m]
                cbeta[n] = cbeta[n] + cc[m] * cbetat[n - m]
    else:
        calpha = calphat
        cbeta = cbetat

    # Computation of integral
    g2 = np.complex(0.0, 0.0)
    del1 = -1.0 + 2.0 * (float(m1) - 1.0) / float(NLS)
    del2 = -1.0 + 2.0 * float(m2) / float(NLS)
    cd1minz = del1 / biglab - z
    cd2minz = del2 / biglab - z
    if abs(cd1minz) < 1.0e-8 / labda:
        cd1minz = cd1minz + 1.0e-8
    if abs(cd2minz) < 1.0e-8 / labda:
        cd2minz = cd2minz + 1.0e-8
    cln1 = np.log(cd1minz)
    cln2 = np.log(cd2minz)
    for n in range(0, 2 * NTERMS - 1 + order + 1):
        g2 = g2 - (calpha[n] * cln2 - calpha[n] / (n + 1) + cbeta[n]) * (cd2minz) ** (
            n + 1
        ) / float(n + 1)
        g2 = g2 + (calpha[n] * cln1 - calpha[n] / (n + 1) + cbeta[n]) * (cd1minz) ** (
            n + 1
        ) / float(n + 1)
    g2 = biglabin * g2 / (2 * np.pi)

    # Integral g3
    # Eq. 23
    calphat[0] = np.complex(0.0, 0.0)
    for n in range(1, 2 * NTERMS):  # Loop start at 1, because of bug in Digital Fortran
        calphat[n] = np.complex(0.0, 0.0)
        for m in range(max(0, n - NTERMS), int((n - 1) / 2) + 1):
            calphat[n] = calphat[n] + cahat[n - m - 1] * cgamma[n - m - 1, m] * (
                -1.0
            ) ** (n - 1 - 2 * m)

    # Compute coefficients of delta^p
    for m in range(0, order + 1):
        cc[m] = RBINOM[order, m] * zbar ** (order - m) * biglab ** order

    if order > 0:
        for n in range(0, 2 * NTERMS + order):
            calpha[n] = np.complex(0.0, 0.0)
            for m in range(max(0, n - 2 * NTERMS + 1), min(n, order) + 1):
                calpha[n] = calpha[n] + cc[m] * calphat[n - m]
    else:
        calpha = calphat

    # Computation of integral
    g3 = np.complex(0.0, 0.0)
    # cd1minz = del1 / biglab - zbar  cd2minz = del2 / biglab - zbar
    # if ( abs(cd1minz) < 1.0e-8) cd1minz = cd1minz + 1.0e-8
    # if ( abs(cd2minz) < 1.0e-8) cd2minz = cd2minz + 1.0e-8
    # By definition log is conjugate of previous log this avoids problems with signs along the line (and saves logs).
    cd1minz = np.conj(cd1minz)
    cd2minz = np.conj(cd2minz)
    cln1 = np.conj(cln1)
    cln2 = np.conj(cln2)
    for n in range(0, 2 * NTERMS + order):
        g3 = g3 - (calpha[n] * cln2 - calpha[n] / (n + 1)) * (cd2minz) ** (
            n + 1
        ) / float(n + 1)
        g3 = g3 + (calpha[n] * cln1 - calpha[n] / (n + 1)) * (cd1minz) ** (
            n + 1
        ) / float(n + 1)
    g3 = biglabin * g3 / (2.0 * np.pi)
    wdis = g1 + g2 + g3
    return wdis


@numba.njit(nogil=True)
def IntegralLapLineDipole(zin, z1, z2, del0, ra, order):
    cg = np.full(order + 2, np.complex(0.0, 0.0))
    z = (2.0 * zin - (z1 + z2)) / (z2 - z1)
    zplus1 = z + 1.0
    zmin1 = z - 1.0

    # We don't always have to do this, so maybe put in condition?
    # Determine coefficients of powers of Delta
    for m in range(0, order + 1):
        cg[m] = RBINOM[order, m] * (-del0) ** (order - m) / ra ** order

    zterm = np.complex(0.0, 0.0)
    for n in range(0, order + 1):
        zterm = zterm + cg[n] * z ** n

    qmtot = np.complex(0.0, 0.0)
    for m in range(1, order + 1):
        qm = np.complex(0.0, 0.0)
        for n in range(1, int((m + 1) / 2) + 1):
            qm = qm + z ** (m - 2 * n + 1) / float(2 * n - 1)
        qmtot = qmtot + 2.0 * cg[m] * qm

    comega = (zterm * np.log(zmin1 / zplus1) + qmtot) / (2.0 * np.pi)
    return comega


@numba.njit(nogil=True)
def IntegralLapLineDipoleDis(zin, z1, z2, del0, ra, order):
    cg = np.full(order + 2, np.complex(0.0, 0.0))

    z = (2.0 * zin - (z1 + z2)) / (z2 - z1)
    zplus1 = z + 1.0
    zmin1 = z - 1.0

    # Determine coefficients of powers of Delta for [ (Delta-Delta_0)/a ] ^ p
    for m in range(0, order + 1):
        cg[m] = RBINOM[order, m] * (-del0) ** (order - m) / ra ** order

    zterm1 = np.complex(0.0, 0.0)
    zterm2 = np.complex(0.0, 0.0)
    for n in range(1, order + 1):
        zterm1 = zterm1 + cg[n] * float(n) * z ** (n - 1)
    for n in range(0, order + 1):
        zterm2 = zterm2 + cg[n] * z ** n

    qmtot = np.complex(0.0, 0.0)
    for m in range(2, order + 1):
        qm = np.complex(0.0, 0.0)
        for n in range(1, int(m / 2) + 1):
            qm = qm + float(m - 2 * n + 1) * z ** (m - 2 * n) / float(2 * n - 1)
        qmtot = qmtot + 2.0 * cg[m] * qm

    wdis = (
        zterm1 * np.log(zmin1 / zplus1) + zterm2 * (1.0 / zmin1 - 1.0 / zplus1) + qmtot
    ) / (2.0 * np.pi)
    return wdis


@numba.njit(nogil=True)
def findm1m2(zin, z1in, z2in, Lin, labda, Rconv):
    # Break integral up in sections of max one labda
    # and find first (m1) and last (m2) section within radius of convergence
    if labda == 0.0:
        NLS = 0
    else:
        NLS = int(np.ceil(Lin / labda))

    m1 = 0
    m2 = 0
    for j in range(1, NLS + 1):
        z1 = z1in + float(j - 1) / NLS * (z2in - z1in)
        z2 = z1 + (z2in - z1in) / NLS
        L = abs(z2 - z1)
        biglab = 2.0 * labda / L
        z = (2.0 * zin - (z1 + z2)) / (z2 - z1) / biglab
        if m1 == 0:
            if abs(z) < Rconv:
                m1 = j
        else:
            if abs(z) > Rconv:
                m2 = j - 1
                break
    if m2 == 0:
        m2 = NLS

    return m1, m2, NLS



import numpy as np

def param_maq(kaq, z, c, npor, top):
    # Computes the parameters for a ModelBase from input for a maq model
    kaq = np.atleast_1d(kaq).astype('d')
    z = np.atleast_1d(z).astype('d')
    c = np.atleast_1d(c).astype('d')
    npor = np.atleast_1d(npor).astype('d')
    if top == 'conf':
        Naq = int(len(z) / 2)
        ltype = np.array( list((Naq-1) * 'al' + 'a' ))
    else: # leaky layer on top
        Naq = int((len(z) - 1) / 2)
        ltype = np.array(list(Naq * 'la'))
    if len(kaq) == 1: kaq = kaq * np.ones(Naq)
    assert len(kaq) == Naq, 'Error: length of kaq needs to be 1 or' + str(Naq)
    H = z[:-1] - z[1:]
    assert np.all(H >= 0), 'Error: Not all layers thicknesses are non-negative' + str(H) 
    if top == 'conf':
        if len(c) == 1: c = c * np.ones(Naq - 1)
        if len(npor) == 1: npor = npor * np.ones(2 * Naq - 1)
        assert len(c) == Naq-1, 'Error: Length of c needs to be 1 or' + str(Naq-1)
        assert len(npor) == 2 * Naq - 1, 'Error: Length of npor needs to be 1 or' + str(2*Naq-1)
        c = np.hstack((1e100,c))
    else: # leaky layer on top
        if len(c) == 1: c = c * np.ones(Naq)
        if len(npor) == 1: npor = npor * np.ones(2 * Naq)
        assert len(c) == Naq, 'Error: Length of c needs to be 1 or' + str(Naq)
        assert len(npor) == 2 * Naq, 'Error: Length of npor needs to be 1 or' + str(2*Naq)
    return kaq, c, npor, ltype

def param_3d(kaq, z, kzoverkh, npor, top='conf', topres=0):
    # Computes the parameters for a ModelBase from input for a model3d model
    kaq = np.atleast_1d(kaq).astype('d')
    z = np.atleast_1d(z).astype('d')
    kzoverkh = np.atleast_1d(kzoverkh).astype('d')
    npor = np.atleast_1d(npor).astype('d')
    if top == 'conf':
        Naq = len(z) - 1
        ltype = np.array(Naq * ['a'])
    elif top == 'semi':
        Naq = len(z) - 1
        ltype = np.hstack(('l', Naq * ['a']))
    if len(kaq) == 1: kaq = kaq * np.ones(Naq)
    assert len(kaq) == Naq, 'Error: length of kaq needs to be 1 or' + str(Naq)
    if len(kzoverkh) == 1: kzoverkh = kzoverkh * np.ones(Naq)
    assert len(kzoverkh) == Naq, 'Error: length of kzoverkh needs to be 1 or' + str(Naq)
    if len(npor) == 1:
        if top == 'conf':
            npor = npor * np.ones(Naq)
        elif top == 'semi':
            npor = npor * np.ones(Naq + 1)
    if top == 'conf':
        assert len(npor) == Naq, 'Error: length of npor needs to be 1 or' + str(Naq)
    elif top == 'semi':
        assert len(npor) == Naq + 1, 'Error: length of npor needs to be 1 or' + str(Naq + 1)
    H = z[:-1] - z[1:]
    assert np.all(H >= 0), 'Error: Not all layers thicknesses are non-negative' + str(H) 
    c = 0.5 * H[:-1] / (kzoverkh[:-1] * kaq[:-1]) + 0.5 * H[1:] / (kzoverkh[1:] * kaq[1:])
    if top == 'conf':
        c = np.hstack((1e100,c))
    elif top == 'semi':
        c = np.hstack((topres + 0.5 * H[0] / (kzoverkh[0] * kaq[0]), c))
    return kaq, c, npor, ltype


import numpy as np
import inspect  # Used for storing the input
from .element import Element
from .equation import DisvecEquation, LeakyWallEquation

__all__ = ['ImpLineDoublet1D', 'LeakyLineDoublet1D']

class LineDoublet1D(Element):

    def __init__(self, model, xld, delp=1, layers=0, name="LineDoublet1DBase", label=None, 
                 addtomodel=True, res=0, aq=None):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers,
                         name=name, label=label)
        self.xld = float(xld)
        self.delp = np.atleast_1d(delp)
        self.res = float(res)
        self.aq = aq
        self.addtomodel = addtomodel
        if self.addtomodel: 
            self.model.add_element(self)
        self.nparam = self.nlayers
        
    def __repr__(self):
        return self.name + " at " + str(self.xld) + " in layers: " + str(self.layers)
    
    def initialize(self):
        self.xc = np.array([self.xld])
        self.yc = np.zeros(1)
        self.ncp = 1
        if self.aq is None:
            self.aq = self.model.aq.find_aquifer_data(self.xc[0], self.yc[0])
        if self.addtomodel:
            self.aq.add_element(self)
        self.parameters = np.empty((self.nparam, 1))
        self.parameters[:, 0] = self.delp
        self.theta_norm_out = np.zeros(1)
        self.cosnorm = np.cos(self.theta_norm_out) * np.ones(self.ncp)
        self.sinnorm = np.sin(self.theta_norm_out) * np.ones(self.ncp)
        self.resfac = self.aq.T[self.layers] / self.res

    def potinf(self, x, y, aq=None):
        if aq is None: 
            aq = self.model.aq.find_aquifer_data(x, 0)
        rv = np.zeros((self.nparam, aq.naq))
        if aq == self.aq:
            pot = np.zeros(aq.naq)
            if aq.ilap:
                if x - self.xld < 0.:
                    #pot[0] = -0.5 * (x - self.xld - 1)  # so that pot = 0.5 at x=xld
                    pot[0] = -0.5   # so that pot = 0.5 at x=xld
                    pot[1:] = -0.5 * np.exp((x - self.xld) / aq.lab[1:])
                elif x - self.xld >= 0.:
                    #pot[0] = 0.5 * (x - self.xld + 1)
                    pot[0] = 0.5
                    pot[1:] = 0.5 * np.exp(-(x - self.xld) / aq.lab[1:])
            else:
                if x - self.xld < 0.:
                    pot[:] = -0.5 * np.exp((x - self.xld) / aq.lab)
                elif x - self.xld >= 0.:
                    pot[:] = 0.5 * np.exp(-(x - self.xld) / aq.lab)
            rv[:] = self.aq.coef[self.layers] * pot
        return rv
    
    def disvecinf(self, x, y, aq=None):
        if aq is None: 
            aq = self.model.aq.find_aquifer_data(x, 0)
        rv = np.zeros((2, self.nparam, aq.naq))
        if aq == self.aq:
            qx = np.zeros(aq.naq)
            if aq.ilap:
                if x - self.xld < 0.:
                    qx[0] = 0.0
                    qx[1:] = 0.5 / aq.lab[1:] * np.exp((x - self.xld) / aq.lab[1:])
                elif x - self.xld >= 0.:
                    qx[0] = 0.0
                    qx[1:] = 0.5 / aq.lab[1:] * np.exp(-(x - self.xld) / aq.lab[1:])
            else:
                if x - self.xld < 0.:
                    qx[:] = 0.5 / aq.lab * np.exp((x - self.xld) / aq.lab)
                elif x - self.xld >= 0.:
                    qx[:] = 0.5 / aq.lab * np.exp(-(x - self.xld) / aq.lab)
            rv[0] = self.aq.coef[self.layers] * qx
        return rv
    
class ImpLineDoublet1D(LineDoublet1D, DisvecEquation):
    """
    Create 1D impermeable wall
    
    """
    
    def __init__(self, model, xld=0, layers=0, label=None):
        self.storeinput(inspect.currentframe())
        LineDoublet1D.__init__(self, model, xld, delp=0, layers=layers, \
                               name="ImpLineDoublet1D", label=label, \
                               addtomodel=True, res=np.inf, aq=None)
        self.nunknowns = self.nparam

    def initialize(self):
        LineDoublet1D.initialize(self)

    def setparams(self, sol):
        self.parameters[:, 0] = sol


class LeakyLineDoublet1D(LineDoublet1D, LeakyWallEquation):
    """
    Create an infinitely long leaky or impermeable wall
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xld : scalar
        x-location of line-doublet
    hls : scalar
        head in line-sink
    res : scalar (default is 0)
        resistance of leaky wall. use np.inf to create impermeable wall 
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element

    """

    tiny = 1e-6

    def __init__(self, model, xld=0, res=np.inf, layers=0, label=None):
        self.storeinput(inspect.currentframe())
        LineDoublet1D.__init__(self, model, xld, delp=0, layers=layers,
                               name="LeakyLineDoublet1D", label=label,
                               addtomodel=True, res=res, aq=None)
        self.nunknowns = self.nparam

    def initialize(self):
        LineDoublet1D.initialize(self)
        self.xcin = self.xc - self.tiny
        self.xcout = self.xc + self.tiny
        self.ycin = self.yc
        self.ycout = self.yc

    def setparams(self, sol):
        self.parameters[:, 0] = sol
